!+
! Module: constants
!
! Authors:  Richard West - SSC (Leicester)
!          
! Description: Define useful mathematical and physical constants
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module constants

  use dal

  implicit none 

 
  public

  ! PI related constants
  real (kind=double), parameter :: Pi = 3.1415926535897931d0
  real (kind=double), parameter :: TwoPi = 2.0d0*Pi
  real (kind=double), parameter :: FourPi = 4.0d0*Pi
  real (kind=double), parameter :: PiOverTwo = 0.5d0*Pi
  real (kind=double), parameter :: PiOn2 = PiOverTwo
  real (kind=double), parameter :: OneOverPi = 1.0d0/Pi
  real (kind=double), parameter :: OneOverTwoPi = 1.0d0/TwoPi

  ! Angular conversion factors
  real (kind=double), parameter :: DegToRad = TwoPi/360.d0
  real (kind=double), parameter :: RadToDeg = 360.d0/TwoPi
  real (kind=double), parameter :: Deg2Rad = DegToRad ! for rgslib/pointing_mod users.
  real (kind=double), parameter :: PiOn180 = DegToRad
  real (kind=double), parameter :: ArcminToRad = DegToRad/60.0d0
  real (kind=double), parameter :: RadToArcmin = RadToDeg*60.0d0
  real (kind=double), parameter :: ArcsecToRad = DegToRad/3600.0d0
  real (kind=double), parameter :: RadToArcsec = RadToDeg*3600.0d0

  ! Solid angle conversion factors
  real (kind=double), parameter :: SqDegToSterad = DegToRad*DegToRad
  real (kind=double), parameter :: SqArcminToSterad = SqDegToSterad/3600.0d0
  real (kind=double), parameter :: SqArcsecToSterad = SqDegToSterad/1.296d7
  real (kind=double), parameter :: SteradToSqDeg = RadToDeg*RadToDeg
  real (kind=double), parameter :: SteradToSqArcmin = SteradToSqDeg*3600.d0
  real (kind=double), parameter :: SteradToSqArcSec = SteradToSqDeg*1.296d7

  ! Square roots
  real (kind=double), parameter :: RootTwo = 1.4142135623730951d0
  real (kind=double), parameter :: RootThree = 1.7320508075688772d0
  real (kind=double), parameter :: RootTen = 3.1622776601683795d0
  real (kind=double), parameter :: RootPi = 1.7724538509055159d0

  ! Natural log
  real (kind=double), parameter :: NapierE = 2.7182818284590455d0
  real (kind=double), parameter :: lnTwo = 0.6931471805599453d0
  real (kind=double), parameter :: lnTen = 2.3025850929940459d0
  real (kind=double), parameter :: lnPi = 1.1447298858494002d0

!  ! Fundamental physical constants (from Zombeck)
!  real (kind=double), parameter :: SpeedOfLight = 299792458.0d0 ! m/sec
!  real (kind=double), parameter :: PlanckH = 6.6260755d-34 ! J/sec
!  real (kind=double), parameter :: NewtonG = 6.67259d-11 ! m^3/kg/s^2

  ! Fundamental physical constants (1998 CODATA recommended)
  real (kind=double), parameter :: SpeedOfLight = 2.99792458d8    ! m/sec
  real (kind=double), parameter :: PlanckH      = 6.62606876d-34  ! J sec
  real (kind=double), parameter :: NewtonG      = 6.673d-11       ! m^3/kg/s^2
  real (kind=double), parameter :: ElectronQ    = 1.602176462d-19 ! coulombs
  real (kind=double), parameter :: Boltzmann    = 1.3806503d-23   ! J/K
  real (kind=double), parameter :: StefanBoltzmann = 5.670400d-8  ! W/m^2/K^4
  real (kind=double), parameter :: Avogadro     = 6.02214199d23   ! mol^-1

  ! Energy conversion factors
!  real (kind=double), parameter :: ErgToeV = 6.24150636309403d11
  real (kind=double), parameter :: ErgToeV = 1.0d-7 / ElectronQ
  real (kind=double), parameter :: ErgTokeV = ErgToeV/1000.0d0
  real (kind=double), parameter :: eVToErg = 1.0d0/ErgToeV
  real (kind=double), parameter :: keVToErg = 1000.0d0/ErgToeV

  ! Solar system constants
  real (kind=double), parameter :: EarthRadius = 6378.14 ! km

contains

  !---------------------------------------------------------------------
  function angstroms2eV(angstroms) result(EeV)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: angstroms
    real(kind(1d0)) :: EeV

    !-------------------------------------------------------------------
    ! Start:

    EeV = PlanckH * SpeedOfLight / (angstroms * 1.0d-10) / ElectronQ

  end function angstroms2eV

  !---------------------------------------------------------------------
  function eV2angstroms(EeV) result(angstroms)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), intent(in) :: EeV
    real(kind(1d0)) :: angstroms

    !-------------------------------------------------------------------
    ! Start:

    angstroms = (PlanckH * SpeedOfLight / EeV / ElectronQ) * 1.0d10

  end function eV2angstroms

  !---------------------------------------------------------------------
  subroutine getBBfluxPerHertz(hertzValues, kelvin, powerValues, status)
  !---------------------------------------------------------------------
  ! This function calculates the power (in watts) radiated per unit
  ! frequency (in hertz) per unit surface area (m^2) per unit solid
  ! angle (sr) at frequency='hertz' by a black body at temperature='kelvin'.
  ! It is designed to be portable and as fast yet accurate as possible.
  ! The precision at low frequency is limited by the calculation of
  ! exp(k1 * f) - 1: for k1 * f less than minArg, a series expansion of
  ! (exp(arg) - 1) / arg is used. The minimum precision occurs at
  ! arg = minArg and is approximately equal to precision(1d0) + log10(arg).
  ! Note that minArg should NOT be greater than 1, otherwise the series
  ! will not converge.


    implicit none

    real(kind(1d0)), intent(in)            :: hertzValues(:), kelvin
    real(kind(1d0)), intent(out)           :: powerValues(:)
    integer,         intent(out), optional :: status

    integer :: localStatus, nValues, i, n
    real(kind(1d0)) :: arg, k1, k2, maxarg, minarg, minTermValue,&
      summ, term

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    powerValues = 0.0d0
    nValues = size(hertzValues)
    if (size(powerValues) .ne. nValues) then
      if (present(status)) status = -1
  return
    end if

    k1 = PlanckH / Boltzmann / kelvin
    k2 = 2.0d0 * PlanckH / SpeedOfLight / SpeedOfLight
    minArg = 0.1d0
    maxArg = log(huge(1d0))
    minTermValue = 10**(-precision(1d0))

    do i = 1, nValues
      arg = k1 * hertzValues(i)
      if (arg .ge. maxArg) then
        powerValues(i) = 0.0d0

      else if (arg .le. minArg) then
        summ = 0.0d0
        term = 1.0d0
        n = 1
        do while(term .gt. minTermValue)
          summ = summ + term
          n = n + 1
          term = term * arg / real(n, kind(1d0))
        end do
        powerValues(i) = k2 * hertzValues(i) * hertzValues(i) * hertzValues(i)&
        / arg / summ

      else
        powerValues(i) = k2 * hertzValues(i) * hertzValues(i) * hertzValues(i)&
        / (exp(arg) - 1.0d0)
      end if
    end do

  end subroutine getBBfluxPerHertz

  !---------------------------------------------------------------------
  function bbFluxPerHertz(hertz, kelvin) result(power)
  !---------------------------------------------------------------------
  ! This function calculates the power (in watts) radiated per unit
  ! frequency (in hertz) per unit surface area (m^2) per unit solid
  ! angle (sr) at frequency='hertz' by a black body at temperature='kelvin'.
  ! It is designed to be portable and as fast yet accurate as possible.
  ! The precision at low frequency is limited by the calculation of
  ! exp(k1 * f) - 1: for k1 * f less than minArg, a series expansion of
  ! (exp(arg) - 1) / arg is used. The minimum precision occurs at
  ! arg = minArg and is approximately equal to precision(1d0) + log10(arg).
  ! Note that minArg should NOT be greater than 1, otherwise the series
  ! will not converge.

    implicit none

    real(kind(1d0)), intent(in) :: hertz, kelvin
    real(kind(1d0)) :: power, arg, k1, k2, maxarg, minarg, minTermValue,&
      summ, term
    logical :: firstCall
    integer :: n

    save firstCall, k1, k2, maxArg, minArg, minTermValue
    data firstCall /.true./
    data minArg /0.1d0/

    !-------------------------------------------------------------------
    ! Start:

!print *, 'in bbFluxPerHertz. hertz, kelvin:', hertz, kelvin
    if (firstCall) then
      k1 = PlanckH / Boltzmann / kelvin
      k2 = 2.0d0 * PlanckH / SpeedOfLight / SpeedOfLight
      maxArg = log(huge(1d0))
      minTermValue = 10**(-precision(1d0))

      firstCall = .false.
!print *, 'first call. k1, k2, maxArg, minTermValue:', k1, k2, maxArg, minTermValue
!    else
!print *, 'subs. call. k1, k2, maxArg, minTermValue:', k1, k2, maxArg, minTermValue
    end if

    arg = k1 * hertz
    if (arg .ge. maxArg) then
      power = 0.0d0

    else if (arg .le. minArg) then
      summ = 0.0d0
      term = 1.0d0
      n = 1
      do while(term .gt. minTermValue)
        summ = summ + term
        n = n + 1
        term = term * arg / real(n, kind(1d0))
      end do
      power = k2 * hertz * hertz * hertz / arg / summ

    else
      power = k2 * hertz * hertz * hertz / (exp(arg) - 1.0d0)
    end if

!print *, 'power:', power
!print *, 'leaving bbFluxPerHertz.'

  end function bbFluxPerHertz

  !---------------------------------------------------------------------
  subroutine getBBfluxPerMetre(angstromValues, kelvin, powerValues, status)
  !---------------------------------------------------------------------
  ! This function calculates the power (in watts) radiated per unit
  ! wavelength (in metres) per unit surface area (m^2) per unit solid
  ! angle (sr) at wavelength='angstroms' by a black body at temperature =
  ! 'kelvin'. It is designed to be portable and as fast yet accurate as
  ! possible. The precision at long wavelengths is limited by the
  ! calculation of exp(k1/lambda) - 1: for k1/lambda less than minArg, a
  ! series expansion of (exp(arg) - 1) / arg is used. The minimum precision
  ! occurs at arg = minArg and is approximately equal to precision(1d0) +
  ! log10(arg). Note that minArg should NOT be greater than 1, otherwise
  ! the series will not converge.

    implicit none

    real(kind(1d0)), intent(in)            :: angstromValues(:), kelvin
    real(kind(1d0)), intent(out)           :: powerValues(:)
    integer,         intent(out), optional :: status

    integer :: localStatus, nValues, i, n
    real(kind(1d0)) :: arg, k1, k2, maxarg, minarg, minTermValue,&
      summ, term

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    powerValues = 0.0d0
    nValues = size(angstromValues)
    if (size(powerValues) .ne. nValues) then
      if (present(status)) status = -1
  return
    end if

    k1 = PlanckH / Boltzmann / kelvin
    k2 = 2.0d0 * PlanckH / SpeedOfLight / SpeedOfLight
    minArg = 0.1d0
    maxArg = log(huge(1d0))
    minTermValue = 10**(-precision(1d0))

    do i = 1, nValues
      arg = k1 / angstromValues(i)
      if (arg .ge. maxArg) then
        powerValues(i) = 0.0d0

      else if (arg .le. minArg) then
        summ = 0.0d0
        term = 1.0d0
        n = 1
        do while(term .gt. minTermValue)
          summ = summ + term
          n = n + 1
          term = term * arg / real(n, kind(1d0))
        end do
        powerValues(i) = k2 / angstromValues(i) / angstromValues(i)&
        / angstromValues(i) / angstromValues(i) / angstromValues(i)&
        / arg / summ

      else
        powerValues(i) = k2 / angstromValues(i) / angstromValues(i)&
        / angstromValues(i) / angstromValues(i) / angstromValues(i)&
        / (exp(arg) - 1.0d0)

      end if
    end do

  end subroutine getBBfluxPerMetre

  !---------------------------------------------------------------------
  function bbFluxPerMetre(angstroms, kelvin) result(power)
  !---------------------------------------------------------------------
  ! This function calculates the power (in watts) radiated per unit
  ! wavelength (in metres) per unit surface area (m^2) per unit solid
  ! angle (sr) at wavelength='angstroms' by a black body at temperature =
  ! 'kelvin'. It is designed to be portable and as fast yet accurate as
  ! possible. The precision at long wavelengths is limited by the
  ! calculation of exp(k1/lambda) - 1: for k1/lambda less than minArg, a
  ! series expansion of (exp(arg) - 1) / arg is used. The minimum precision
  ! occurs at arg = minArg and is approximately equal to precision(1d0) +
  ! log10(arg). Note that minArg should NOT be greater than 1, otherwise
  ! the series will not converge.

    implicit none

    real(kind(1d0)), intent(in) :: angstroms, kelvin
    real(kind(1d0)) :: power, arg, k1, k2, maxarg, minarg, minTermValue,&
      summ, term
    logical :: firstCall
    integer :: n

    save firstCall, k1, k2, maxArg, minArg, minTermValue
    data firstCall /.true./
    data minArg /0.1d0/

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) then
      k1 = PlanckH * SpeedOfLight / 1.0d-10 / Boltzmann / kelvin
      k2 = 2.0d0 * PlanckH * SpeedOfLight * SpeedOfLight
      maxArg = log(huge(1d0))
      minTermValue = 10**(-precision(1d0))

      firstCall = .false.
    end if

    arg = k1 / angstroms
    if (arg .ge. maxArg) then
      power = 0.0d0

    else if (arg .le. minArg) then
      summ = 0.0d0
      term = 1.0d0
      n = 1
      do while(term .gt. minTermValue)
        summ = summ + term
        n = n + 1
        term = term * arg / real(n, kind(1d0))
      end do
      power = k2 / angstroms / angstroms / angstroms / angstroms / angstroms&
      / arg / summ

    else
      power = k2 / angstroms / angstroms / angstroms / angstroms / angstroms&
      / (exp(arg) - 1.0d0)
    end if

  end function bbFluxPerMetre
end module constants
