!
! Module:dss_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains auxiliary routines to do with the data subspace (DSS).
! 
! Comments: TODO:
! 
! $Id: dss_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dss_aux

  use dal
  use errorHandling
  use dss

  use fits_utils
  use ssc_misc
  use sort_mod

  implicit none 

  private 

  public loadBoxChain
  public calcBoxChainOverlaps
  private calcBoxChainOverlap
  private calcBoxOverlap
  private dissectBox
  public copyRange
  public copyRanges
  private orLists
  private copyBoxChain

  public loadRanges
  private dssCompObeys
  private dssFilterOverlaps
  public andAllRanges
  public orAllRanges
  public checkRangeOverlap
  public andRanges
  private orRanges
  private xorRanges
  public correctRanges
  public rangesAreValid
  private swapRangeBounds
  private sortRanges
  private sortBounds
  private swapBounds
  private quickSortStatic
  private rangesAreIdentical
  private filterTypeToStr
  public rangeTypeToStr
  public numDssCompsOfAxis
  public numDssCompsWithValue
  public getNum1stDssCompWithValue
  public anyFilterOfThisAxis
  public numFiltersOfAxis
  public get1stFilterThisAxis
  public allFiltersPassValue
  public valueIsPassedByComp
  public valueIsPassedByDssFilter
  public valueIsWithinRange
  public dumpDss
  public filterSetIsIdentical
  public filtersAreIdentical
  private decomposeXyNameString
  public getRangeSpan
  public dumpRangeChain
  public dumpBoxChain
  public valuesArePassedByBoxChain
  public projectBoxChainToRangeChain

  interface loadRanges
    module procedure loadRangesDss
    module procedure loadRangesComp
  end interface

  interface dumpDss
    module procedure dumpDssComponent
    module procedure dumpDssBlock
    module procedure dumpDssPointer
  end interface

  type, public :: dssConstraintType
    character(20)         :: name
    logical               :: ignore
    type(RangeT), pointer :: ranges(:)
  end type dssConstraintType
  type, public :: rangeBoxType
    type(RangeT) :: x, y
  end type rangeBoxType
  type, public :: boxChainType
    type(rangeBoxType), pointer :: boxes(:)
    integer(int32), pointer :: comps(:)
    character(20) :: xAxisName, yAxisName
  end type boxChainType
  type, public :: dssCompBlockType
    type(RangeT), pointer :: rangeChain(:)
    integer, pointer :: compNums(:)
  end type dssCompBlockType

  integer, public, parameter :: BISTOOLOW = 1, BISTOOHIGH = 2, BOVERLAPSA = 3

contains 

  !---------------------------------------------------------------------
  subroutine loadBoxChain(xRange, yRange, compNumBoolVector&
  , xAxisName, yAxisName, boxChain)
  !---------------------------------------------------------------------
  ! This takes 2 sets of ranges (supposedly orthogonal) and returns them as a single structure vector boxChain(). The two sets of ranges define a sort of checkerboard region in x-y space, ie made up of intersecting rectangles. If there are N xRanges and M yRanges then clearly there will be N*M intersection boxes. Each element of boxChain() contains the details of one of these boxes. The order of boxes in boxChain is not significant.
  !
  ! Oh yes, boxChain also points to another vector which contains the component numbers associated with the ranges.
  !
  ! NOTE that boxChain()%boxes and %comps are pointers and so should be deallocated after use by the calling routine.

    implicit none

    type(RangeT),        intent(in)  :: xRange(:), yRange(:)
    logical,             intent(in)  :: compNumBoolVector(:)
    type(boxChainType),  intent(out) :: boxChain
    character(*),        intent(in)  :: xAxisName, yAxisName

    integer(int16) :: xsize, ysize, numComps, boxNum, xi, yi, compNum, i

    !-------------------------------------------------------------------
    ! Start

    xsize = size(xRange)
    ysize = size(yRange)
    numComps = count(compNumBoolVector)
    allocate(boxChain%boxes(xsize * ysize), boxChain%comps(numComps))
    boxNum = 0
    do xi = 1, xsize
      do yi = 1, ysize
        boxNum = boxNum + 1
        boxChain%boxes(boxNum)%x = copyRange(xRange(xi))
        boxChain%boxes(boxNum)%y = copyRange(yRange(yi))
      end do
    end do

    compNum = 0
    do i = 1, size(compNumBoolVector)
    if (.not. compNumBoolVector(i)) cycle
      compNum = compNum + 1
      boxChain%comps(compNum) = i
    end do

    boxChain%xAxisName = xAxisName
    boxChain%yAxisName = yAxisName

  end subroutine loadBoxChain

  !---------------------------------------------------------------------
  subroutine calcBoxChainOverlaps(boxChains, newBoxChain)
  !---------------------------------------------------------------------
  ! The purpose of this subroutine is to examine how the boxes in the single chain 'newBoxChain' overlap those of all the box chains in 'boxChains'. The effect may be both to resize 'boxChains' and, where necessary, resize or alter its constituent chains, so as to (i) include new chains comprising the overlaps between all the input chains; (ii) the original chains but with all overlaps clipped or removed from their constituent boxes; (iii) a new chain comprising the chain input in 'newBoxChain', but with all overlap regions clipped from its boxes.
  !
  ! Note that the pointer arrays newBoxChain%boxes and newBoxChain%comps
  ! are deallocated in this routine.

    implicit none

    type(boxChainType), pointer :: boxChains(:)
    type(boxChainType), intent(inout) :: newBoxChain

    integer(int32) :: numInChains, numOutChains, inChainNum, outChainNum
    type(boxChainType) :: overlapChain
    type(boxChainType), pointer :: tempBoxChains(:)

    !-------------------------------------------------------------------
    ! Start

    numInChains = size(boxChains)
    numOutChains = 0

    do inChainNum = 1, numInChains
    if (size(newBoxChain%boxes) <= 0) exit ! no point in checking for
      ! more overlaps since newBoxChain is now entirely eaten away.

      call calcBoxChainOverlap(boxChains(inChainNum), newBoxChain&
      , overlapChain)

      if (size(boxChains(inChainNum)%boxes) > 0) then
        numOutChains = numOutChains + 1
        call copyBoxChain(boxChains(inChainNum)&
          , tempBoxChains(numOutChains))
      end if

      if (size(overlapChain%boxes) > 0) then ! there were some overlaps.
        numOutChains = numOutChains + 1
        call copyBoxChain(overlapChain, tempBoxChains(numOutChains))
      end if

      deallocate(boxChains(inChainNum)%boxes)
      deallocate(boxChains(inChainNum)%comps)
    end do

    if (size(newBoxChain%boxes) > 0) then
      numOutChains = numOutChains + 1
      call copyBoxChain(newBoxChain, tempBoxChains(numOutChains))
    end if

    allocate(boxChains(numOutChains))
    do outChainNum = 1, numOutChains
      call copyBoxChain(tempBoxChains(numOutChains)&
        , boxChains(numOutChains))
    end do

  end subroutine calcBoxChainOverlaps

  !---------------------------------------------------------------------
  subroutine calcBoxChainOverlap(inChain1, inChain2, overlapChain)
  !---------------------------------------------------------------------
  ! NOTE! the returned pointers %boxes and %comps in inChain1, inChain2
  ! and overlapChain should be deallocated after use.
  !
  ! The input arguments contain two elements, 'boxes' and 'comps', both of them vector pointer arrays. The elements of 'boxes' describe rectangular regions which represent a combination of DSS filters on two axes. The array 'comps' gives a set of DSS component numbers each of which has a combination of filters on these two axes which is fully compatible with that stored in 'boxes' - that is, DSS components which have filters on the axes which, if ANDed with that in 'boxes', would return the identical filter as that in 'boxes'.
  !
  ! The purpose of the subroutine is to examine how the boxes in inChain1 overlap those of inChain2. Where there is any overlap, the chain of boxes describing the overlap regions is returned in the variable overlapChain. overlapChain%comps is in this case set to contain the union of inChain1%comps and inChain2%comps, since these overlap regions are consistent now with both sets of DSS components, under the definition given above. If there is no overlap, both overlapChain%boxes and overlapChain%comps are allocated to size = 0.
  !
  ! In case of some overlap occurring, the input chains are 'clipped' to exclude the overlap regions. This may entail some resizing of the 'boxes' element of either or both (in extreme cases, where one is entirely contained within the other for example, to 0). This comes about as follows.
   !
   ! The subroutine checks for overlap between each pair of rectangles of the two input chains. Now, the overlap region between two rectangles is itself always a rectangle, so the generation of a new element of the overlap chain is easy to manage, but as already said we want also to clip these overlap regions from the input rectangles. However, the only way to represent a rectangle with a bite taken out of it is by replacing it with two abutting, different sized rectangles. Eg:
    !
    !	aaaaaaaaaaaaaaaaaaaaaaaaaa
    !	a                        a
    !	a                        a
    !	a              11111111111111111
    !	a              1         a     1
    !	a              1         a     1
    !	a              1         a     1
    !	a              1         a     1
    !	aaaaaaaaaaaaaaa1aaaaaaaaaa     1
    !	               1               1
    !	               1               1
    !	               1               1
    !	               11111111111111111
    !
    ! should become
    !
    !	cccccccccccccccccccccccccc
    !	c                        c
    !	cccccccccccccccccccccccccc
    !	bbbbbbbbbbbbbbbAAAAAAAAAAA222222
    !	b             bA         A2    2
    !	b             bA         A2    2
    !	b             bA         A2    2
    !	b             bA         A2    2
    !	bbbbbbbbbbbbbbbAAAAAAAAAAA222222
    !	               33333333333333333
    !	               3               3
    !	               3               3
    !	               33333333333333333
    !
    ! Here 'b' and 'c' comprise the clipped return of input chain 'a', '2' and '3' comprise what remains of input '1', and 'A' represents the overlap.
    !
    ! Note that there is ambiguity in the way in which a clipped input rectangle is divided up: the line of division can be either horizontal (separation in the y coordinate) or vertical (separation in the x coordinate). The tossed coin came down for horizontal, so that is the way it is done herein.

    implicit none

    type(boxChainType), intent(inout) :: inChain1, inChain2
    type(boxChainType), intent(out) :: overlapChain

    integer(int32) ::&
      numBoxes1, numBoxes2, guessAtMaxSize, maxSize1, maxSize2,&
      maxSizeOverlap, numClippedBoxes1, numClippedBoxes2,&
      numOverlapBoxes, boxNum1, boxNum2, box2loopLimit, numExtra,&
      start1, finis1, start2, finis2, i
    type(rangeBoxType), allocatable ::&
      tempClippedBoxes1(:), tempClippedBoxes2(:), tempOverlapBoxes(:)
    type(rangeBoxType), pointer ::&
      extraBoxFragments1(:), extraBoxFragments2(:)
    type(rangeBoxType) ::&
      overlapBox
    logical ::&
      thereWasOverlap
    logical, allocatable ::&
      box1HasBeenErased(:), box2HasBeenErased(:), checkArray(:,:)

    !-------------------------------------------------------------------
    ! Start

    numBoxes1 = size(inChain1%boxes)
    numBoxes2 = size(inChain2%boxes)

    guessAtMaxSize = 5 * (numBoxes1 + numBoxes2)
    maxSize1 = guessAtMaxSize
    maxSize2 = guessAtMaxSize
    maxSizeOverlap = guessAtMaxSize
    allocate(tempClippedBoxes1(maxSize1), box1HasBeenErased(maxSize1))
    allocate(tempClippedBoxes2(maxSize2), box2HasBeenErased(maxSize2))
    allocate(tempOverlapBoxes(maxSizeOverlap))
    allocate(checkArray(maxSize1, maxSize2))

    tempClippedBoxes1(:numBoxes1) = inChain1%boxes
    tempClippedBoxes2(:numBoxes2) = inChain2%boxes
    checkArray = .false.
    numClippedBoxes1 = numBoxes1
    numClippedBoxes2 = numBoxes2
    numOverlapBoxes = 0
    box1HasBeenErased = .false.
    box2HasBeenErased = .false.

    boxNum1 = 1
    do while(boxNum1 <= numClippedBoxes1)
    if (box1HasBeenErased(boxNum1)) cycle

      box2loopLimit = numClippedBoxes2
      do boxNum2 = 1, box2loopLimit ! It is true that the number
        ! of tempClippedBoxes2 can increase during this loop, in
        ! exactly the same way as the number of tempClippedBoxes1
        ! can in the outer loop, so I could for the sake of
        ! consistency have made the inner loop exit the same type
        ! as that of the outer loop. However I have taken a
        ! slight shortcut because there is no point in testing
        ! any of the boxes2 added during the inner loop because
        ! they are all the fragments of boxes2 from which the
        ! overlap against the current box1 has already been clipped.

      if (box2HasBeenErased(boxNum2)) cycle

      if (checkArray(boxNum1, boxNum2)) cycle ! can't be any overlap with this
      ! combo, don't bother to check.

        call calcBoxOverlap(tempClippedBoxes1(boxNum1)&
          , tempClippedBoxes2(boxNum2), extraBoxFragments1&
          , extraBoxFragments2, thereWasOverlap, box1HasBeenErased(boxNum1)&
          , box2HasBeenErased(boxNum2), overlapBox)

        ! If there is clipping, the returned values of
        ! tempClippedBoxes(boxNum) are clipped as appropriate and
        ! any extra fragments returned in extraBoxFragments.

        checkArray(boxNum1, boxNum2) = .true.

        if (thereWasOverlap) then
          numOverlapBoxes = numOverlapBoxes + 1
!** test if bigger than maxSizeOverlap and increase if so

          tempOverlapBoxes(numOverlapBoxes) = overlapBox

          numExtra = size(extraBoxFragments1)
          if (numExtra > 0) then
            numClippedBoxes1 = numClippedBoxes1 + numExtra
!** test if bigger than maxSize1 and increase if so

            start1 = numClippedBoxes1 - numExtra + 1
            finis1 = numClippedBoxes1
            tempClippedBoxes1(start1:finis1) = extraBoxFragments1

            ! These new boxes1 are the children of the current box1.
            ! Copy the checks of the latter to the former:
            checkArray(start1:finis1, 1:numClippedBoxes2)&
              = spread(checkArray(boxNum1, 1:numClippedBoxes2), 1, numExtra)
          end if

          numExtra = size(extraBoxFragments2)
          if (numExtra > 0) then
            numClippedBoxes2 = numClippedBoxes2 + numExtra
!** test if bigger than maxSize2 and increase if so

            start2 = numClippedBoxes2 - numExtra + 1
            finis2 = numClippedBoxes2
            tempClippedBoxes2(start2:finis2) = extraBoxFragments2

            ! These new boxes2 are the children of the current box2.
            ! Copy the checks of the latter to the former:
            checkArray(1:numClippedBoxes1, start2:finis2)&
              = spread(checkArray(1:numClippedBoxes1, boxNum2), 2, numExtra)
          end if
        end if ! there was overlap.
      end do

      boxNum1 = boxNum1 + 1
    end do

    deallocate(extraBoxFragments1, extraBoxFragments2)
    deallocate(inChain1%boxes, inChain2%boxes)
!    allocate(inChain1%boxes(numClippedBoxes1))
!    allocate(inChain2%boxes(numClippedBoxes2))
allocate(inChain1%boxes(count(.not. box1HasBeenErased(:numClippedBoxes1))))
allocate(inChain2%boxes(count(.not. box2HasBeenErased(:numClippedBoxes2))))
    allocate(overlapChain%boxes(numOverlapBoxes))

!    inChain1%boxes = tempClippedBoxes1(:numClippedBoxes1)
!    inChain2%boxes = tempClippedBoxes2(:numClippedBoxes2)
i = 0
do boxNum1 = 1, numClippedBoxes1
if (box1HasBeenErased(boxNum1)) cycle
  i = i + 1
  inChain1%boxes(i) = tempClippedBoxes1(boxNum1)
end do
i = 0
do boxNum2 = 1, numClippedBoxes2
if (box2HasBeenErased(boxNum2)) cycle
  i = i + 1
  inChain2%boxes(i) = tempClippedBoxes2(boxNum2)
end do
    overlapChain%boxes = tempOverlapBoxes(:numOverlapBoxes)

    deallocate(tempClippedBoxes1, tempClippedBoxes2, tempOverlapBoxes)
    deallocate(box1HasBeenErased, box2HasBeenErased, checkArray)

    if (size(overlapChain%boxes) > 0) then ! some of the boxes overlapped.
      overlapChain%comps = orLists(inChain1%comps, inChain2%comps)
    end if

  end subroutine calcBoxChainOverlap

  !---------------------------------------------------------------------
  subroutine calcBoxOverlap(box1, box2, extraBoxes1&
  , extraBoxes2, thereisOverlap, box1HasBeenErased, box2HasBeenErased&
  , overlapBox)
  !---------------------------------------------------------------------
  ! NOTE! the returned pointers extraBoxes1 and extraBoxes2 should
  ! be deallocated after use.

    implicit none

    type(rangeBoxType), intent(inout) :: box1, box2
    type(rangeBoxType), pointer :: extraBoxes1(:), extraBoxes2(:)
    logical, intent(out) :: thereisOverlap, box1HasBeenErased, box2HasBeenErased
    type(rangeBoxType), intent(out) :: overlapBox

    !-------------------------------------------------------------------
    ! Start

    box1HasBeenErased = .false. ! default
    box2HasBeenErased = .false. ! default

    ! Check first to see if there is any overlap at all.

    if   (checkRangeOverlap(box1%x, box2%x) == BOVERLAPSA&
    .and. checkRangeOverlap(box1%y, box2%y) == BOVERLAPSA) then
      thereisOverlap = .true.
    else
      thereisOverlap = .false.
    end if

    if (thereisOverlap) then
      overlapBox%x = andRanges(box1%x, box2%x)
      overlapBox%y = andRanges(box1%y, box2%y)

      call dissectBox(overlapBox, box1, extraBoxes1, box1HasBeenErased)
      call dissectBox(overlapBox, box2, extraBoxes2, box2HasBeenErased)

    else
      allocate(extraBoxes1(0), extraBoxes2(0))
    end if

  end subroutine calcBoxOverlap

  !---------------------------------------------------------------------
  subroutine dissectBox(overlapBox, box, extraBoxes, boxHasBeenErased)
  !---------------------------------------------------------------------
  ! NOTE! the returned pointer extraBoxes should be deallocated after use.
  !
  ! This routine IS NOT GUARANTEED TO WORK unless there is some overlap
  ! between the boxes (ie, overlapBox has sensible values). This should
  ! therefore be tested beforehand.

    implicit none

    type(rangeBoxType), intent(in) :: overlapBox
    type(rangeBoxType), intent(inout) :: box
    type(rangeBoxType), pointer :: extraBoxes(:)
    logical, intent(out) :: boxHasBeenErased

    integer(int32) :: numXORedX, numXORedY, i
    type(RangeT), pointer :: xoredRangesX(:), xoredRangesY(:)

    !-------------------------------------------------------------------
    ! Start

    call xorRanges(box%x, overlapBox%x, xoredRangesX)
    call xorRanges(box%y, overlapBox%y, xoredRangesY)
    numXORedX = size(xoredRangesX)
    numXORedY = size(xoredRangesY)
    if (numXORedX < 1) then
      if (numXORedY < 1) then
        boxHasBeenErased = .true.
        allocate(extraBoxes(0))

      else ! numXORedY > 0
        box%y = copyRange(xoredRangesY(1))
        ! unnecessary to change box%x value.

        allocate(extraBoxes(numXORedY - 1))
        do i = 2, numXORedY
          extraBoxes(i - 1)%x = copyRange(box%x)
          extraBoxes(i - 1)%y = copyRange(xoredRangesY(i))
        end do
      end if

    else ! numXORedX > 0
      if (numXORedY < 1) then
        box%x = copyRange(xoredRangesX(1))
        ! unnecessary to change box%y value.

        allocate(extraBoxes(numXORedX - 1))
        do i = 2, numXORedX
          extraBoxes(i - 1)%x = copyRange(xoredRangesX(i))
          extraBoxes(i - 1)%y = copyRange(box%y)
        end do

      else ! numXORedY > 0
        box%y = copyRange(xoredRangesY(1))
        ! unnecessary to change box%x value.

        allocate(extraBoxes(numXORedX + numXORedY - 1))
        do i = 2, numXORedY
          extraBoxes(i - 1)%x = copyRange(box%x)
          extraBoxes(i - 1)%y = copyRange(xoredRangesY(i))
        end do
        do i = numXORedY + 1, numXORedY + numXORedX
          extraBoxes(i - 1)%x = copyRange(xoredRangesX(i - numXORedY))
          extraBoxes(i - 1)%y = copyRange(overlapBox%y)
        end do
      end if
    end if

    deallocate(xoredRangesX, xoredRangesY)

  end subroutine dissectBox

  !---------------------------------------------------------------------
  function copyRange(inRange) result(outRange)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: inRange
    type(RangeT) :: outRange

    !-------------------------------------------------------------------
    ! Start

    outRange%lower%type  = inRange%lower%type
    outRange%lower%value = inRange%lower%value
    outRange%upper%type  = inRange%upper%type
    outRange%upper%value = inRange%upper%value

  end function copyRange

  !---------------------------------------------------------------------
  function orLists(inList1, inList2) result(oredList)
  !---------------------------------------------------------------------
!*** put it in ssc_misc??

  ! The returned pointer oredList should be deallocated after use.
  !
  ! The function should NOT be called as in the following example:
  !
  !	oredList => orLists(oredList, list2)
  !
  ! The memory that the pointer oredList pointed to before the call now
  ! has no pointer to it, since at the moment the call is executed
  ! oredList is reassigned to the same memory that the function points
  ! to, which was newly assigned during the call. The initial memory
  ! pointed to by oredList therefore cannot now be deallocated. Better
  ! would be:
  !
  !	tempOredList => orLists(oredList, list2)
  !	deallocate(oredList)
  !	oredList => tempOredList
  !
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: inList1(:), inList2(:)
    integer(int32), pointer :: oredList(:)

    integer(int32) :: tempOredList(size(inList1) + size(inList2)),&
      size1, size2, numOred, i

    !-------------------------------------------------------------------
    ! Start

    size1 = size(inList1)
    size2 = size(inList2)
    tempOredList(1:size1) = inList1
    numOred = size1
    do i = 1, size2
      if (.not. any(tempOredList(:numOred) == inList2(i))) then
        numOred = numOred + 1
        tempOredList(numOred) = inList2(i)
      end if
    end do

    allocate(oredList(numOred))
    oredList = tempOredList(:numOred)

  end function orLists

  !---------------------------------------------------------------------
  subroutine copyBoxChain(inChain, outChain)
  !---------------------------------------------------------------------
  ! Copies the pointer arrays in the input to the output and
  ! then deallocates the input.
  !
  ! NOTE that the pointers %boxes and %comps in outChain should be
  ! deallocated after use.

    implicit none

    type(boxChainType), intent(inout) :: inChain
    type(boxChainType), intent(out) :: outChain

    integer(int32) :: numBoxes, numComps

    !-------------------------------------------------------------------
    ! Start

    numBoxes = size(inChain%boxes)
    numComps = size(inChain%comps)
    allocate(outChain%boxes(numBoxes))
    allocate(outChain%comps(numComps))
    outChain%boxes = inChain%boxes
    outChain%comps = inChain%comps
    deallocate(inChain%boxes, inChain%comps)

  end subroutine copyBoxChain

  !---------------------------------------------------------------------
  subroutine loadRangesDss(dssPointer, axisName, rangeChain&!, status&
  , dssConstraints)
  !---------------------------------------------------------------------
  ! The function of this subroutine is to return a vector of dss type
  ! RangeT containing the ranges of the axis 'axisName'. Note that the sub
  ! will succeed only if that subset of the set of DSS components which obeys the supplied constraints can be expressed as (ranges in axisName) OR (all filters on other axes). This has been a complicated job to implement because a chain of ranges is implicitely an OR of the individual ranges and for this reason can be distributed among several components. The subroutine works therefore as follows. Firstly, blocks of components are identified which each include all components obeying the constraints and having the same range chain in axisName. Secondly, the non-axisName filters of each block of components are checked for identity. If they all prove to be the same, then their respective range chains are ORed together.
  !
  ! The returned pointer rangeChain should be deallocated after use.

    implicit none

    type(DssT),              intent(in)            :: dssPointer
    character(*),            intent(in)            :: axisName
    type(RangeT),            pointer               :: rangeChain(:)
    type(dssConstraintType), intent(in), optional  :: dssConstraints(:)

    logical :: thereAreConstraints, noCompObeysConstraints, isNewRangeChain
    integer :: numComps, compNum, numBlocks, blockNum
    type(dssCompBlockType), pointer :: blockChain(:)
    type(DScompT) :: dssComp
    type(RangeT), pointer :: compRangeChain(:), tempRangeChn(:)
    character(20) :: axisNamesToNeglect(1)

    !-------------------------------------------------------------------
    ! Start

    thereAreConstraints = .false.
    if (present(dssConstraints)) then
      if (size(dssConstraints) > 0) thereAreConstraints = .true.
    end if

    numComps = dssComponents(dssPointer)
    if (numComps .le. 0) then
!*** status = ???
      allocate(rangeChain(0))
  return
    end if

    noCompObeysConstraints = .true. ! default
    do compNum = 0, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      noCompObeysConstraints = .false.
!      call loadRangesComp(dssComp, axisName, firstRanges)
!      startCompNum = compNum + 1
    exit
    end do

    if (noCompObeysConstraints) then
!*** status = ???
      allocate(rangeChain(0))
  return
    end if

    !...................................................................
    ! Start of main routine. First load all the blocks of constraint-
    ! -passing components:

    numBlocks = 0
    allocate(blockChain(0))
    do compNum = 0, numComps - 1
!print *, ''
!print *, 'compNum', compNum
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      call loadRangesComp(dssComp, axisName, compRangeChain)

      ! Does it have a range chain the same as any of the previous range chains?
      isNewRangeChain = .true. ! default
      do blockNum = 1, numBlocks
        if (rangeChainsIdentical(compRangeChain&
          , blockChain(blockNum)%rangeChain)) then

          isNewRangeChain = .false.
        end if
      end do

      if (isNewRangeChain) then ! make a new block
        call addNewBlock(blockChain, numBlocks, compRangeChain, compNum)
!print *, '  block added. numBlocks now =', numBlocks
      else ! it is the same rangeChain as a previous block
        call addNewCompNum(blockChain(numBlocks)%compNums, compNum)
!print *, '  range chain same as previous block.'

      end if

      deallocate(compRangeChain)
    end do
if (numBlocks <= 0) call error('bug','')

    !...................................................................
    ! Now check that, apart from filters on axisName, the blocks are the same:

    axisNamesToNeglect(1) = axisName
    do blockNum = 2, numBlocks
      if (.not. componentBlocksIdentical(dssPointer, blockChain(blockNum)&
        , blockChain(1), axisNamesToNeglect)) then

        allocate(rangeChain(0))
!** status??
  return
      end if
    end do

    !...................................................................
    ! OR together the total range chain:

    allocate(rangeChain(size(blockChain(1)%rangeChain)))
    rangeChain = blockChain(1)%rangeChain
    do blockNum = 2, numBlocks
      tempRangeChn => orRangeChains(blockChain(blockNum)%rangeChain, rangeChain)
      deallocate(rangeChain)
      rangeChain => tempRangeChn
    end do

    !...................................................................
    ! Clear up all the block pointers:

    do blockNum = 1, numBlocks
      deallocate(blockChain(blockNum)%compNums, blockChain(blockNum)%rangeChain)
    end do
    deallocate(blockChain)

  end subroutine loadRangesDss

  !---------------------------------------------------------------------
  subroutine addNewBlock(compBlockChain, numCompBlocks, rangeChain, compNum)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompBlockType), pointer       :: compBlockChain(:)
    integer,             intent(inout) :: numCompBlocks
    type(RangeT),        intent(in)    :: rangeChain(:)
    integer,             intent(in)    :: compNum

    type(dssCompBlockType), pointer :: tempCompBlockChain(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempCompBlockChain(numCompBlocks))
    call copyComponentBlockChain(compBlockChain, tempCompBlockChain)
    deallocate(compBlockChain)
    numCompBlocks = numCompBlocks + 1
    allocate(compBlockChain(numCompBlocks))
    call copyComponentBlockChain(tempCompBlockChain&
      , compBlockChain(:numCompBlocks-1))
    deallocate(tempCompBlockChain)
    call copyRanges(rangeChain, compBlockChain(numCompBlocks)%rangeChain)
    allocate(compBlockChain(numCompBlocks)%compNums(1))
    compBlockChain(numCompBlocks)%compNums(1) = compNum

  end subroutine addNewBlock

  !---------------------------------------------------------------------
  subroutine copyComponentBlockChain(inCompBlockChain, outCompBlockChain)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompBlockType), intent(in)  :: inCompBlockChain(:)
    type(dssCompBlockType), intent(out) :: outCompBlockChain(size(inCompBlockChain))

    integer :: numBlocks, blockNum

    !-------------------------------------------------------------------
    ! Start

    numBlocks = size(inCompBlockChain)
    do blockNum = 1, numBlocks
      outCompBlockChain(blockNum)&
        = copyComponentBlock(inCompBlockChain(blockNum))
    end do

  end subroutine copyComponentBlockChain

  !---------------------------------------------------------------------
  function copyComponentBlock(inCompBlock) result(outCompBlock)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompBlockType), intent(in) :: inCompBlock
    type(dssCompBlockType)             :: outCompBlock

    !-------------------------------------------------------------------
    ! Start

    call copyRanges(inCompBlock%rangeChain, outCompBlock%rangeChain)
    allocate(outCompBlock%compNums(size(inCompBlock%compNums)))
    outCompBlock%compNums = inCompBlock%compNums

  end function copyComponentBlock

  !---------------------------------------------------------------------
  subroutine copyRanges(inRangeChain, outRangeChain)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: inRangeChain(:)
    type(RangeT), pointer    :: outRangeChain(:)

    integer :: numRanges, rangeNum

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(inRangeChain)
    allocate(outRangeChain(numRanges))
    do rangeNum = 1, numRanges
      outRangeChain(rangeNum) = copyRange(inRangeChain(rangeNum))
    end do

  end subroutine copyRanges

  !---------------------------------------------------------------------
  subroutine addNewCompNum(compNums, compNum)
  !---------------------------------------------------------------------

    implicit none

    integer, pointer    :: compNums(:)
    integer, intent(in) :: compNum

    integer :: numComps
    integer, allocatable :: tempCompNums(:)

    !-------------------------------------------------------------------
    ! Start

    numComps = size(compNums)
    allocate(tempCompNums(numComps))
    tempCompNums = compNums
    deallocate(compNums)
    numComps = numComps + 1
    allocate(compNums(numComps))
    compNums(:numComps-1) = tempCompNums
!    deallocate(tempCompNums)
    compNums(numComps) = compNum

  end subroutine addNewCompNum

  !---------------------------------------------------------------------
  logical function componentBlocksIdentical(dssPointer, componentBlockA&
  , componentBlockB, axisNamesToNeglect)
  !---------------------------------------------------------------------
!**** Not finished. axisNamesToNeglect should be optional (also should include optional axisNamesToInclude), with some (graceful hopefully!) way of passing them to componentsIdentical().

    implicit none

    type(DssT),             intent(in) :: dssPointer
    type(dssCompBlockType), intent(in) :: componentBlockA,&
                                          componentBlockB
    character(*),           intent(in) :: axisNamesToNeglect(:)

    integer :: numComps, compNumA, compNumB
    logical, allocatable :: blockBCompMatched(:)
    logical :: unableToMatchComponent

    !-------------------------------------------------------------------
    ! Start

    componentBlocksIdentical = .false. ! default

    numComps = size(componentBlockA%compNums)
  if (numComps .ne. size(componentBlockB%compNums)) return

    allocate(blockBCompMatched(numComps))
    blockBCompMatched = .false. ! initially.
    do compNumA = 1, numComps
      unableToMatchComponent = .true. ! default
      do compNumB = 1, numComps
      if (blockBCompMatched(compNumB)) cycle
        if (componentsIdentical(dssPointer, componentBlockA%compNums(compNumA)&
          , componentBlockB%compNums(compNumB), axisNamesToNeglect)) then
          unableToMatchComponent = .false.
          blockBCompMatched(compNumB) = .true.
      exit
        end if
      end do

  if (unableToMatchComponent) return
    end do

  if (any(.not. blockBCompMatched)) return

    componentBlocksIdentical = .true. ! made it through all the tests.

  end function componentBlocksIdentical

  !---------------------------------------------------------------------
  logical function componentsIdentical(dssPointer, compNumA, compNumB&
  , axisNamesToInclude, axisNamesToNeglect)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in)           :: dssPointer
    integer,      intent(in)           :: compNumA, compNumB
    character(*), intent(in), optional :: axisNamesToInclude(:),&
                                          axisNamesToNeglect(:)

    type(DScompT) :: dssCompA, dssCompB

    !-------------------------------------------------------------------
    ! Start

    dssCompA = dssComponent(dssPointer, compNumA)
    dssCompB = dssComponent(dssPointer, compNumB)

!*** DON'T forget comp numbers start at ZERO

    if (present(axisNamesToInclude)) then
      if (present(axisNamesToNeglect)) then
        componentsIdentical = componentsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToInclude, axisNamesToNeglect)

      else
        componentsIdentical = componentsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToInclude)
      end if

    else
      if (present(axisNamesToNeglect)) then
        componentsIdentical = componentsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToNeglect=axisNamesToNeglect)

      else
        componentsIdentical = componentsIdenticalComp(dssCompA , dssCompB)
      end if
    end if

  end function componentsIdentical

  !---------------------------------------------------------------------
  logical function componentsIdenticalComp(dssCompA, dssCompB&
  , axisNamesToInclude, axisNamesToNeglect)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT), intent(in)           :: dssCompA, dssCompB
    character(*),  intent(in), optional :: axisNamesToInclude(:),&
                                           axisNamesToNeglect(:)

    integer :: numFiltersA, numFiltersB, filtNumA, filtNumB
    logical, allocatable :: compBFilterMatched(:)
    character(20) :: filterNameA, filterNameB
    logical :: unableToMatchFilter
    type(DSfilterT) :: filterA, filterB

    !-------------------------------------------------------------------
    ! Start

    componentsIdenticalComp = .false. ! default

    numFiltersA = dssFilters(dssCompA)
    numFiltersB = dssFilters(dssCompA)

    allocate(compBFilterMatched(numFiltersB))
    compBFilterMatched = .false. ! initially.
    do filtNumA = 1, numFiltersA
      ! Skip over filterA if it either _isn't_ in the list of axis names
      ! to include, or it _is_ in the list of axis names to neglect.

      filterNameA = dssFilterName(dssFilter(dssCompA, filtNumA-1))
      if (present(axisNamesToInclude)) then
    if (.not. any(filterNameA .eq. axisNamesToInclude)) cycle
      end if
      if (present(axisNamesToNeglect)) then
    if (any(filterNameA .eq. axisNamesToNeglect)) cycle
      end if

      unableToMatchFilter = .true. ! default
      do filtNumB = 1, numFiltersB
      if (compBFilterMatched(filtNumB)) cycle

        filterNameB = dssFilterName(dssFilter(dssCompB, filtNumB-1))
      if (filterNameA .ne. filterNameB) cycle

        ! Tick filterB off if it either _isn't_ in the list of axis names
        ! to include, or it _is_ in the list of axis names to neglect.

        if (present(axisNamesToInclude)) then
          if (.not. any(filterNameB .eq. axisNamesToInclude))&
            compBFilterMatched(filtNumB) = .true.
        end if
        if (present(axisNamesToNeglect)) then
          if (any(filterNameB .eq. axisNamesToNeglect))&
            compBFilterMatched(filtNumB) = .true.
        end if

        filterA = dssFilter(dssCompA, filtNumA - 1)
        filterB = dssFilter(dssCompB, filtNumB - 1)
        if (filtersAreIdentical(filterA, filterB)) then
          unableToMatchFilter = .false.
          compBFilterMatched(filtNumB) = .true.
      exit
        end if
      end do

  if (unableToMatchFilter) return
    end do

  if (any(.not. compBFilterMatched)) return

    componentsIdenticalComp = .true. ! made it through all the tests.

  end function componentsIdenticalComp

  !---------------------------------------------------------------------
  subroutine loadRangesDss_20030314(inBlock, axisName, ranges, status, dssConstraints)
  !---------------------------------------------------------------------
  ! The function of this subroutine is to return a vector of dss type
  ! RangeT containing the ranges of the axis 'axisName'. Note that the sub
  ! will succeed only if the range in each component which obeys the
  ! constraints is the same. Eg 1: if dssConstraints is omitted entirely, ALL
  ! components must have the same range of 'axisName', otherwise the routine
  ! then returns an error. Eg 2: suppose dssConstraints has only 1 element,
  ! of value
  !
  !	%name = 'CCDNR'
  !	%ranges =
  !		%upper = 4.5
  !		%lower = 3.5
  !
  ! In this case, only those components for which CCDNR==4 are examined.
  ! All of these must have the same ranges for 'axisName', otherwise an error
  ! is returned.

    implicit none

    type(BlockT),            intent(in)           :: inBlock
    character(*),            intent(in)           :: axisName
    type(RangeT),         pointer              :: ranges(:)
    integer,                 intent(out)          :: status
    type(dssConstraintType), intent(in), optional :: dssConstraints(:)

    type(DssT)      :: dssPointer
    type(DScompT)   :: dssComp
    integer         :: numComps, compNum, startCompNum
    logical         :: thereAreConstraints, noCompObeysConstraints
    type(RangeT), pointer :: firstRanges(:), nextRanges(:)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    thereAreConstraints = .false.
    if (present(dssConstraints)) then
      if (size(dssConstraints) .gt. 0) thereAreConstraints = .true.
    end if

    dssPointer = dataSubspace(inBlock)

    numComps = dssComponents(dssPointer)
    if (numComps .le. 0) then
!*** status = ???
      allocate(ranges(0))
  return
    end if

    noCompObeysConstraints = .true. ! default
    do compNum = 0, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      noCompObeysConstraints = .false.
      call loadRangesComp(dssComp, axisName, firstRanges)
      startCompNum = compNum + 1
    exit
    end do

    if (noCompObeysConstraints) then
!*** status = ???
      allocate(ranges(0))
  return
    end if

    do compNum = startCompNum, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      call loadRangesComp(dssComp, axisName, nextRanges)

      if (.not. rangesAreIdentical(firstRanges, nextRanges)) then
!*** status = ???
        allocate(ranges(0))
        deallocate(firstRanges, nextRanges)
  return
      end if
    end do

    allocate(ranges(size(firstRanges)))
    ranges = firstRanges
    deallocate(firstRanges, nextRanges)

  end subroutine loadRangesDss_20030314

  !---------------------------------------------------------------------
  subroutine loadRangesComp(dssComp, axisName, compRangeChain)
  !---------------------------------------------------------------------
  ! NOTE that the pointer compRangeChain() should be deallocated after use.

    type(DScompT), intent(in)  :: dssComp
    character(*),  intent(in)  :: axisName
    type(RangeT), pointer  :: compRangeChain(:)

!    logical :: firstFilterOnThisAxis
    integer :: filtNum
    type(DSfilterT) :: filter
    type(RangeT),  pointer :: rangeChain(:), tempRangeChain(:)
    character(20) :: dssFilterTypeStr

    !-------------------------------------------------------------------
    ! Start

    allocate(compRangeChain(1))
    compRangeChain%lower%type = UNDEFINED
    compRangeChain%upper%type = UNDEFINED
    compRangeChain%lower%value = 0.0 ! unnecessary, but I don't
    compRangeChain%upper%value = 0.0 ! like to leave them hanging.

  if (numFiltersOfAxis(dssComp, axisName) .le. 0) return

    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
    if (dssFilterName(filter) .ne. axisName) cycle

      if (dssFilterType(filter) .eq. RANGEFILTERT) then
        rangeChain => dssRanges(filter)
        tempRangeChain => andAllRanges(compRangeChain, rangeChain)
        deallocate(compRangeChain)
        deallocate(rangeChain)
        compRangeChain => tempRangeChain

      else
        allocate(compRangeChain(0))
        dssFilterTypeStr = filterTypeToStr(dssFilterType(filter))
        call warning('badFilterType', "A filter of type "&
          //trim(dssFilterTypeStr)//" on the axis "//trim(axisName)&
          //" was detected. I couldn't read the information from this: &
          &I can only read the information in RANGEFILTERT filters.")
      end if
    end do

  end subroutine loadRangesComp

  !---------------------------------------------------------------------
  subroutine loadRangesComp_20021031(dssComp, axisName, compRanges)
  !---------------------------------------------------------------------
  ! NOTE that the pointer compRanges() should be deallocated after use.

    type(DScompT), intent(in)  :: dssComp
    character(*),  intent(in)  :: axisName
    type(RangeT), pointer  :: compRanges(:)

    logical :: firstFilterOnThisAxis
    integer :: filtNum
    type(DSfilterT) :: filter
    type(RangeT),  pointer :: ranges(:)
    character(20) :: dssFilterTypeStr

    !-------------------------------------------------------------------
    ! Start

    if (numFiltersOfAxis(dssComp, axisName) .le. 0) then
      allocate(compRanges(1))
      compRanges%lower%type = UNDEFINED
      compRanges%upper%type = UNDEFINED
      compRanges%lower%value = 0.0 ! unnecessary, but I don't
      compRanges%upper%value = 0.0 ! like to leave them hanging.
  return
    end if

    firstFilterOnThisAxis = .true.
    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
    if (dssFilterName(filter) .ne. axisName) cycle

      if (dssFilterType(filter) .eq. RANGEFILTERT) then
        if (firstFilterOnThisAxis) then
          firstFilterOnThisAxis = .false.
!          ranges => dssRanges(filter)
!          compRanges = oldToNew(ranges)
compRanges => dssRanges(filter)
          deallocate(ranges)

        else
          ranges => dssRanges(filter)
!          compRanges = andAllRanges(compRanges, oldToNew(ranges))
compRanges => andAllRanges(compRanges, ranges)
          deallocate(ranges)
        end if

      else
        allocate(compRanges(0))
        dssFilterTypeStr = filterTypeToStr(dssFilterType(filter))
        call warning('badFilterType', "A filter of type "&
          //trim(dssFilterTypeStr)//" on the axis "//trim(axisName)&
          //" was detected. I couldn't read the information from this: &
          &I can only read the information in RANGEFILTERT filters.")
      end if
    end do

  end subroutine loadRangesComp_20021031

  !---------------------------------------------------------------------
  logical function dssCompObeys(dssComp, dssConstraints)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT),           intent(in) :: dssComp
    type(dssConstraintType), intent(in) :: dssConstraints(:)

    integer :: numConstraints, i, filtNum
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    dssCompObeys = .true. ! default

    numConstraints = size(dssConstraints)
    do i = 1, numConstraints
    if (numFiltersOfAxis(dssComp, dssConstraints(i)%name) .le. 0) cycle
      do filtNum = 0, dssFilters(dssComp) - 1
        filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .ne. dssConstraints(i)%name) cycle

        if (dssFilterOverlaps(filter, dssConstraints(i))) then
          dssCompObeys = .false.
  return
        end if
      end do
    end do

  end function dssCompObeys

  !---------------------------------------------------------------------
  logical function rangesAreIdentical(rangesA, rangesB)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangesA(:), rangesB(:)

    integer :: numRanges, i

    !-------------------------------------------------------------------
    ! Start

    rangesAreIdentical = .false. ! default

  if (size(rangesA) .ne. size(rangesB)) return
    numRanges = size(rangesA)
    do i = 1, numRanges
      if (rangesA(i)%lower%type .ne. rangesB(i)%lower%type&
      .or. rangesA(i)%upper%type .ne. rangesB(i)%upper%type)&
  return

      if (rangesA(i)%lower%type .ne. UNDEFINED&
      .and. rangesA(i)%lower%value .ne. rangesB(i)%lower%value)&
  return

      if (rangesA(i)%upper%type .ne. UNDEFINED&
      .and. rangesA(i)%upper%value .ne. rangesB(i)%upper%value)&
  return
    end do

    rangesAreIdentical = .true.

  end function rangesAreIdentical

  !---------------------------------------------------------------------
  logical function rangeChainsIdentical(rangeChainA, rangeChainB)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeChainA(:), rangeChainB(:)

    integer :: numRanges, i

    !-------------------------------------------------------------------
    ! Start

    rangeChainsIdentical = .false. ! default

!print *, 'size(rangeChainA), size(rangeChainB):', size(rangeChainA), size(rangeChainB)

  if (size(rangeChainA) .ne. size(rangeChainB)) return
    numRanges = size(rangeChainA)
    do i = 1, numRanges
!print *, 'i:', i

!print *, 'rangeChainA(i)%lower%type, rangeChainB(i)%lower%type:'&
!, rangeChainA(i)%lower%type, rangeChainB(i)%lower%type

!print *, 'rangeChainA(i)%upper%type, rangeChainB(i)%upper%type:'&
!, rangeChainA(i)%upper%type, rangeChainB(i)%upper%type

!print *, 'rangeChainA(i)%lower%value, rangeChainB(i)%lower%value:'&
!, rangeChainA(i)%lower%value, rangeChainB(i)%lower%value

!print *, 'rangeChainA(i)%upper%value, rangeChainB(i)%upper%value:'&
!, rangeChainA(i)%upper%value, rangeChainB(i)%upper%value

      if (rangeChainA(i)%lower%type .ne. rangeChainB(i)%lower%type&
      .or. rangeChainA(i)%upper%type .ne. rangeChainB(i)%upper%type)&
  return

      if (rangeChainA(i)%lower%type .ne. UNDEFINED&
      .and. rangeChainA(i)%lower%value .ne. rangeChainB(i)%lower%value)&
  return

      if (rangeChainA(i)%upper%type .ne. UNDEFINED&
      .and. rangeChainA(i)%upper%value .ne. rangeChainB(i)%upper%value)&
  return
    end do

    rangeChainsIdentical = .true.

  end function rangeChainsIdentical

  !---------------------------------------------------------------------
  logical function dssFilterOverlaps(filter, dssConstraint)
  !---------------------------------------------------------------------

    implicit none

    type(DSfilterT),         intent(in) :: filter
    type(dssConstraintType), intent(in) :: dssConstraint

    type(RangeT), pointer :: ranges(:)
    type(RangeT), pointer :: newRanges(:)

    !-------------------------------------------------------------------
    ! Start

    dssFilterOverlaps = .true. ! default

  if (dssFilterType(filter) .ne. RANGEFILTERT) return
  if (size(dssConstraint%ranges) <= 0) return

    ranges => dssRanges(filter)
    if (size(ranges) <= 0) then
      deallocate(ranges)
  return
    end if

!    newRanges = andAllRanges(oldToNew(ranges), dssConstraint%ranges)
    newRanges => andAllRanges(ranges, dssConstraint%ranges)
    if (size(newRanges) <= 0) dssFilterOverlaps = .false.

    deallocate(ranges, newRanges)

  end function dssFilterOverlaps

  !---------------------------------------------------------------------
  function andAllRanges(rangesA, rangesB, doChecks) result(andedRanges)
  !---------------------------------------------------------------------
!*** check if can get rid of localRanges.

  ! The returned pointer andedRanges should be deallocated after use.
  !
  ! The function should NOT be called as in the following example:
  !
  !	andedRanges => andAllRanges(andedRanges, rangesB)
  !
  ! The memory that the pointer andedRanges pointed to before the call now
  ! has no pointer to it, since at the moment the call is executed
  ! andedRanges is reassigned to the same memory that the function points
  ! to, which was newly assigned during the call. The initial memory
  ! pointed to by andedRanges therefore cannot now be deallocated. Better
  ! would be:
  !
  !	tempAndedRanges => andAllRanges(andedRanges, rangesB)
  !	deallocate(andedRanges)
  !	andedRanges => tempAndedRanges
  !
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangesA(:), rangesB(:)
    type(RangeT), pointer    :: andedRanges(:)
    logical, intent(in), optional :: doChecks

    type(RangeT) ::&
      tempRanges(size(rangesA) + size(rangesB))
    type(RangeT), pointer ::&
      localRangesA(:), localRangesB(:)
    integer :: numRangesA, numRangesB, andedRangeNum, startB, i, j

    !-------------------------------------------------------------------
    ! Start

    allocate(localRangesA(size(rangesA)))
    localRangesA = rangesA
    allocate(localRangesB(size(rangesB)))
    localRangesB = rangesB

    if (present(doChecks)) then
      if (doChecks) then
        ! Enforce correct ordering and overlap on ranges:
        call correctRanges(localRangesA)
        call correctRanges(localRangesB)
      end if
    end if

    numRangesA = size(localRangesA)
    numRangesB = size(localRangesB)

    if (numRangesA .le. 0 .or. numRangesB .le. 0) then
      deallocate(localRangesA, localRangesB)
      allocate(andedRanges(0))
  return
    end if

! At this point both localRangesA and localRangesB obey the following
! rules (provided that doChecks was set):
!	(i) For all i, if both ranges(i)%lower%type and ranges(i)%upper%type are
! not UNDEFINED, then ranges(i)%lower%value < ranges(i)%upper%value, unless both
! lo%type and hi%type are INCLUSIVE, in which case ranges(i)%lower%value may
! be equal to ranges(i)%upper%value; in other words, all ranges include some
! values.
!	(ii) For all i > 1 and <= numRanges, no lo%type may be UNDEFINED.
! In other words, the only lo%type which is allowed to be UNDEFINED is the
! first.
!	(iii) For all i >= 1 and < numRanges, no hi%type may be UNDEFINED.
! In other words, the only hi%type which is allowed to be UNDEFINED is the
! last.
!	(iv) For all i >= 1 and < numRanges, ranges(i)%upper%value <
! ranges(i+1)%lower%value, unless both ranges(i)%upper%type and ranges(i+1)%lower%type
! are EXCLUSIVE, in which case ranges(i)%upper%value may be equal to
! ranges(i+1)%lower%value. In other words, there must be some values between
! adjacent ranges.

    ! Check to see if either range includes (or both ranges include) the whole of the real number axis:

    if (any(localRangesA%lower%type .eq. UNDEFINED&
    .and. localRangesA%upper%type .eq. UNDEFINED)) then
      if (any(localRangesB%lower%type .eq. UNDEFINED&
      .and. localRangesB%upper%type .eq. UNDEFINED)) then
        deallocate(localRangesA, localRangesB)
        allocate(andedRanges(1))
        andedRanges(1)%lower%type = UNDEFINED
        andedRanges(1)%upper%type = UNDEFINED
  return

      else
        allocate(andedRanges(numRangesB))
        andedRanges = localRangesB
        deallocate(localRangesA, localRangesB)
  return
      end if

    else if (any(rangesB%lower%type .eq. UNDEFINED&
    .and. rangesB%upper%type .eq. UNDEFINED)) then
      allocate(andedRanges(numRangesA))
      andedRanges = localRangesA
      deallocate(localRangesA, localRangesB)
  return
    end if

    ! Now compute the AND. This is done as follows. We go through one of the
    ! two sequences of intervals (it doesn't matter which) one by one. For
    ! each interval, we go through a subset of the other sequence and check
    ! for overlap. Action is then taken depending on the result of this check.

    andedRangeNum = 0
    startB = 1
!*** work out startA and maybe finisA also.
    do i = 1, numRangesA
      BLOOP: do j = startB, numRangesB
        select case(checkRangeOverlap(localRangesA(i), localRangesB(j)))
        case(BISTOOHIGH) ! The B range (and all the following ones) are too
        ! high in value to overlap this A range. Quit looking at Bs and
        ! try the next A.

          startB = max(1, j - 1) ! because this must be the first B tested
          ! which has gone over. The previous one must therefore either
          ! have overlapped the A range or been below it (or, there were
          ! no previous Bs, this is the very first). In either case, the
          ! previous B (if there was one) is a convenient start point next
          ! time around.
      exit BLOOP

        case(BISTOOLOW) ! The B range is too low in value to overlap this A
        ! range. Try the next B.
      cycle BLOOP
        end select

        ! Otherwise, BOVERLAPSA, the two ranges must overlap. Let us deal
        ! with that then.

        andedRangeNum = andedRangeNum + 1
        tempRanges(andedRangeNum) = andRanges(localRangesA(i), localRangesB(j))
      end do BLOOP
    end do

    deallocate(localRangesA, localRangesB)

    allocate(andedRanges(andedRangeNum))
    andedRanges = tempRanges(:andedRangeNum)

  end function andAllRanges

  !---------------------------------------------------------------------
  function orAllRanges(rangeChainA, rangeChainB)&
  result(oredRangeChain)
  !---------------------------------------------------------------------
!*** change name to orRangeChains?

  ! The returned pointer oredRangeChain should be deallocated after use.
  !
  ! The function should NOT be called as in the following example:
  !
  !	oredRangeChain => orAllRanges(oredRangeChain, rangeChainB)
  !
  ! The memory that the pointer oredRangeChain pointed to before the call now
  ! has no pointer to it, since at the moment the call is executed
  ! oredRangeChain is reassigned to the same memory that the function points
  ! to, which was newly assigned during the call. The initial memory
  ! pointed to by oredRangeChain therefore cannot now be deallocated. Better
  ! would be:
  !
  !	tempOredRangeChain => orAllRanges(oredRangeChain, rangeChainB)
  !	deallocate(oredRangeChain)
  !	oredRangeChain => tempOredRangeChain
  !
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeChainA(:), rangeChainB(:)
    type(RangeT), pointer    :: oredRangeChain(:)

    !-------------------------------------------------------------------
    ! Start

    ! Simplest way to do this is to simply join the two range chains
    ! together and send them to correctRanges.

    allocate(oredRangeChain(size(rangeChainA) + size(rangeChainB)))
    oredRangeChain(:size(rangeChainA))&
      = rangeChainA
    oredRangeChain(size(rangeChainA)+1:size(rangeChainA)+size(rangeChainB))&
      = rangeChainB

    call correctRanges(oredRangeChain)

  end function orAllRanges

  !---------------------------------------------------------------------
  function orRangeChains(rangeChainA, rangeChainB)&
  result(oredRangeChain)
  !---------------------------------------------------------------------
  ! The returned pointer oredRangeChain should be deallocated after use.
  !
  ! The function should NOT be called as in the following example:
  !
  !	oredRangeChain => orRangeChains(oredRangeChain, rangeChainB)
  !
  ! The memory that the pointer oredRangeChain pointed to before the call now
  ! has no pointer to it, since at the moment the call is executed
  ! oredRangeChain is reassigned to the same memory that the function points
  ! to, which was newly assigned during the call. The initial memory
  ! pointed to by oredRangeChain therefore cannot now be deallocated. Better
  ! would be:
  !
  !	tempOredRangeChain => orRangeChains(oredRangeChain, rangeChainB)
  !	deallocate(oredRangeChain)
  !	oredRangeChain => tempOredRangeChain
  !
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeChainA(:), rangeChainB(:)
    type(RangeT), pointer    :: oredRangeChain(:)

    !-------------------------------------------------------------------
    ! Start

    ! Simplest way to do this is to simply join the two range chains
    ! together and send them to correctRanges.

    allocate(oredRangeChain(size(rangeChainA) + size(rangeChainB)))
    oredRangeChain(:size(rangeChainA))&
      = rangeChainA
    oredRangeChain(size(rangeChainA)+1:size(rangeChainA)+size(rangeChainB))&
      = rangeChainB

    call correctRanges(oredRangeChain)

  end function orRangeChains

!  !---------------------------------------------------------------------
!  function orOneWithManyRanges(oredRangeChain, range)&
!  result(newOredRangeChain)
!  !---------------------------------------------------------------------
!
!  ! The function should NOT be called as in the following example:
!  !
!  !	oredRangeChain => orOneWithManyRanges(oredRangeChain, range)
!  !
!  ! The memory that the pointer oredRanges pointed to before the call now
!  ! has no pointer to it, since oredRanges was reassociated during the call.
!  ! This memory cannot now be deallocated. Better would be:
!  !
!  !	tempOredRangeChain => orOneWithManyRanges(oredRangeChain, range)
!  !	deallocate(oredRangeChain)
!  !	oredRangeChain => tempOredRangeChain
!  !
!  ! Remember to deallocate oredRangeChain after use.
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    type(RangeT), intent(in) :: oredRangeChain(:), range
!    type(RangeT), pointer    :: newOredRangeChain(:)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!!    if () then
!!    else
!!    end if
!
!  end function orOneWithManyRanges

  !---------------------------------------------------------------------
  function checkRangeOverlap(rangeA, rangeB) result(status)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeA, rangeB
    integer :: status

    !-------------------------------------------------------------------
    ! Start

    if (rangeB%lower%type .ne. UNDEFINED&
    .and. (rangeB%lower%value .gt. rangeA%upper%value)&
    .or. rangeA%upper%type .eq. EXCLUSIVE&
    .and. rangeB%lower%type .eq. EXCLUSIVE&
    .and. (rangeB%lower%value .ge. rangeA%upper%value)) then
    ! The B range is too high in value to overlap the A range.
      status = BISTOOHIGH

    else if (rangeB%upper%type .ne. UNDEFINED&
    .and. (rangeB%upper%value .lt. rangeA%lower%value)&
    .or. rangeA%lower%type .eq. EXCLUSIVE&
    .and. rangeB%lower%type .eq. EXCLUSIVE&
    .and. (rangeB%lower%value .ge. rangeA%upper%value)) then
    ! The B range is too low in value to overlap the A range.
      status = BISTOOLOW

    else
      status = BOVERLAPSA
    end if

  end function checkRangeOverlap

  !---------------------------------------------------------------------
  logical function rangesAreValid(ranges)
  !---------------------------------------------------------------------
  ! This checks that the range vector (of size n) obeys the following conditions:
  !	(i) Only ranges(1)%lower%type and ranges(n)%upper%type are permitted to be UNDEFINED.
  !	(ii) For all ranges for which neither lower or upper type = UNDEFINED (ie, for all internal ranges), upper%value > lower%value unless both types are INCLUSIVE, in which case upper%value may equal lower%value. This means that each range must encompass some non-zero set of real numbers. 
  !	(iii) For all ranges, ranges(i)%upper%value < ranges(i+1)%lower%value unless both ranges(i)%upper%type and ranges(i+1)%lower%type are EXCLUSIVE, in which case ranges(i)%upper%value may equal ranges(i+1)%lower%value. This means that the gap between two adjacent ranges must encompass some non-zero set of real numbers.
  !
  ! Note that any ranges pointer can be brought into valid condition by passing it through the subroutine correctRanges().

    implicit none

    type(RangeT), intent(in) :: ranges(:)
    integer :: numRanges

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(ranges)
    if (numRanges <= 0) then
      rangesAreValid = .true.
  return
    end if

    rangesAreValid = .false. ! default

    if (numRanges > 1) then
  if (any(ranges(2:           )%lower%type == UNDEFINED)) return
  if (any(ranges( :numRanges-1)%upper%type == UNDEFINED)) return

  if (any(ranges(2:)%upper%value > ranges(:numRanges-1)%lower%value)) return

      if (any(ranges(2:)%upper%value >= ranges(:numRanges-1)%lower%value&
      .and. ranges(2:           )%upper%type == INCLUSIVE&
      .and. ranges( :numRanges-1)%lower%type == INCLUSIVE))&
  return
    end if

    if (any(ranges%upper%value < ranges%lower%value&
    .and. ranges%upper%type .ne. UNDEFINED&
    .and. ranges%lower%type .ne. UNDEFINED))&
  return

    if (any(ranges%upper%value <= ranges%lower%value&
    .and. ranges%upper%type == EXCLUSIVE&
    .and. ranges%lower%type == EXCLUSIVE))&
  return

  rangesAreValid = .true. ! passed all the tests.

  end function rangesAreValid

  !---------------------------------------------------------------------
  function andRanges(rangeA, rangeB) result(andedRanges)
  !---------------------------------------------------------------------
  ! WARNING! This function will not work UNLESS THE RANGES OVERLAP as tested by
  ! checkRangeOverlap().
  !
  ! The returned pointer andedRanges should be deallocated after use.
  !
  ! The function should NOT be called as in the following example:
  !
  !	andedRanges => andRanges(andedRanges, rangeB)
  !
  ! The memory that the pointer andedRanges pointed to before the call now
  ! has no pointer to it, since at the moment the call is executed
  ! andedRanges is reassigned to the same memory that the function points
  ! to, which was newly assigned during the call. The initial memory
  ! pointed to by andedRanges therefore cannot now be deallocated. Better
  ! would be:
  !
  !	tempAndedRanges => andRanges(andedRanges, rangeB)
  !	deallocate(andedRanges)
  !	andedRanges => tempAndedRanges
  !
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeA, rangeB
    type(RangeT) :: andedRanges

    !-------------------------------------------------------------------
    ! Start

    ! First calculate the value and type of the start of the ANDed range:
    if (rangeA%lower%type .eq. UNDEFINED) then
      if (rangeB%lower%type .eq. UNDEFINED) then
        andedRanges%lower%type = UNDEFINED
        ! Value doesn't matter.

      else
        andedRanges%lower  = rangeB%lower
      end if

    else
      if (rangeB%lower%type .eq. UNDEFINED) then
        andedRanges%lower  = rangeA%lower 
      else ! neither rangeA%lower%type nor rangeB%lower%type is UNDEFINED.
        if      (rangeA%lower%value .gt. rangeB%lower%value) then
          andedRanges%lower  = rangeA%lower 
        else if (rangeB%lower%value .gt. rangeA%lower%value) then
          andedRanges%lower  = rangeB%lower 
        else ! rangeA%lower%value = rangeB%lower%value.
          andedRanges%lower%value = rangeA%lower%value ! ist egal.

          if (rangeA%lower%type .eq. EXCLUSIVE&
          .or. rangeB%lower%type .eq. EXCLUSIVE) then ! we must exclude the
          ! boundary point.
            andedRanges%lower%type = EXCLUSIVE
          else
            andedRanges%lower%type = INCLUSIVE
          end if
        end if
      end if
    end if

    ! Now calculate the value and type of the end of the ANDed range:
    if (rangeA%upper%type .eq. UNDEFINED) then
      if (rangeB%upper%type .eq. UNDEFINED) then
        andedRanges%upper%type = UNDEFINED
        ! Value doesn't matter.

      else
        andedRanges%upper  = rangeB%upper 
      end if

    else
      if (rangeB%upper%type .eq. UNDEFINED) then
        andedRanges%upper  = rangeA%upper 
      else ! neither rangeA%upper%type nor rangeB%upper%type is UNDEFINED.
        if      (rangeA%upper%value .lt. rangeB%upper%value) then
          andedRanges%upper  = rangeA%upper 
        else if (rangeB%upper%value .lt. rangeA%upper%value) then
          andedRanges%upper  = rangeB%upper 
        else ! rangeA%upper%value = rangeB%upper%value.
          andedRanges%upper%value = rangeA%upper%value ! ist egal.

          if (rangeA%upper%type .eq. EXCLUSIVE&
          .or. rangeB%upper%type .eq. EXCLUSIVE) then ! we must exclude the
          ! boundary point.
            andedRanges%upper%type = EXCLUSIVE
          else
            andedRanges%upper%type = INCLUSIVE
          end if
        end if
      end if
    end if

  end function andRanges

  !---------------------------------------------------------------------
  function orRanges(rangeA, rangeB) result(oredRanges)
  !---------------------------------------------------------------------
  ! WARNING! This function will not work UNLESS THE RANGES OVERLAP as tested by
  ! checkRangeOverlap().
  !
  ! The returned pointer oredRanges should be deallocated after use.
  !
  ! The function should NOT be called as in the following example:
  !
  !	oredRanges => orRanges(oredRanges, rangeB)
  !
  ! The memory that the pointer oredRanges pointed to before the call now
  ! has no pointer to it, since at the moment the call is executed
  ! oredRanges is reassigned to the same memory that the function points
  ! to, which was newly assigned during the call. The initial memory
  ! pointed to by oredRanges therefore cannot now be deallocated. Better
  ! would be:
  !
  !	tempOredRanges => orRanges(oredRanges, rangeB)
  !	deallocate(oredRanges)
  !	oredRanges => tempOredRanges
  !
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeA, rangeB
    type(RangeT) :: oredRanges

    !-------------------------------------------------------------------
    ! Start

    ! First calculate the value and type of the start of the ORed range:
    if (rangeA%lower%type .eq. UNDEFINED .or. rangeB%lower%type .eq. UNDEFINED)&
    then
      oredRanges%lower%type = UNDEFINED
      ! Value doesn't matter.

    else ! neither rangeA%lower%type nor rangeB%lower%type is UNDEFINED.
      if      (rangeA%lower%value .lt. rangeB%lower%value) then
        oredRanges%lower  = rangeA%lower
      else if (rangeB%lower%value .lt. rangeA%lower%value) then
        oredRanges%lower  = rangeB%lower 
      else ! rangeA%lower%value = rangeB%lower%value.
        oredRanges%lower%value = rangeA%lower%value ! ist egal.

        if (rangeA%lower%type .eq. INCLUSIVE&
        .or. rangeB%lower%type .eq. INCLUSIVE) then ! we must include the
        ! boundary point.
          oredRanges%lower%type = INCLUSIVE
        else
          oredRanges%lower%type = INCLUSIVE
        end if
      end if
    end if

    ! Now calculate the value and type of the end of the ORed range:
    if (rangeA%upper%type .eq. UNDEFINED .or. rangeB%upper%type .eq. UNDEFINED)&
    then
      oredRanges%upper%type = UNDEFINED
      ! Value doesn't matter.

    else ! neither rangeA%upper%type nor rangeB%upper%type is UNDEFINED.
      if      (rangeA%upper%value .gt. rangeB%upper%value) then
        oredRanges%upper  = rangeA%upper
      else if (rangeB%upper%value .gt. rangeA%upper%value) then
        oredRanges%upper  = rangeB%upper 
      else ! rangeA%upper%value = rangeB%upper%value.
        oredRanges%upper%value = rangeA%upper%value ! ist egal.

        if (rangeA%upper%type .eq. INCLUSIVE&
        .or. rangeB%upper%type .eq. INCLUSIVE) then ! we must include the
        ! boundary point.
          oredRanges%upper%type = INCLUSIVE
        else
          oredRanges%upper%type = EXCLUSIVE
        end if
      end if
    end if

  end function orRanges

  !---------------------------------------------------------------------
  subroutine xorRanges(range, rangeSubset, xoredRanges)
  !---------------------------------------------------------------------
  ! This returns those fragments (if any) of range that are _not_ part
  ! of rangeSubset.
  !
  ! The pointer xoredRanges should be deallocated after use.

!*** maybe make it more general, so that it returns xor of both input ranges??

    implicit none

    type(RangeT), intent(in) :: range, rangeSubset
    type(RangeT), pointer :: xoredRanges(:)

    logical :: noOverlap, lowerEndRemnantExists, upperEndRemnantExists
    integer(int8) :: numFragments, fragmentNum

    !-------------------------------------------------------------------
    ! Start

    ! Have to accommodate the trivial case in which there is no overlap
    ! at all between range and rangeSubset:

    noOverlap = .false. ! default
    if (range%lower%type .ne. UNDEFINED&
    .and. rangeSubset%upper%type .ne. UNDEFINED) then
      if (range%lower%type == INCLUSIVE) then
        if  ((rangeSubset%upper%type == INCLUSIVE&
        .and. rangeSubset%upper%value <  range%lower%value)&
        .or. (rangeSubset%upper%type == EXCLUSIVE&
        .and. rangeSubset%upper%value <= range%lower%value))&
          noOverlap = .true.

      else ! range%lower%type == EXCLUSIVE
        if (rangeSubset%upper%value <= range%lower%value)&
          noOverlap = .true.

      end if

    else if (range%upper%type .ne. UNDEFINED&
    .and. rangeSubset%lower%type .ne. UNDEFINED) then
      if (range%upper%type == INCLUSIVE) then
        if  ((rangeSubset%lower%type == INCLUSIVE&
        .and. rangeSubset%lower%value >  range%upper%value)&
        .or. (rangeSubset%lower%type == EXCLUSIVE&
        .and. rangeSubset%lower%value >= range%upper%value))&
          noOverlap = .true.

      else ! range%upper%type == EXCLUSIVE
        if (rangeSubset%lower%value >= range%upper%value)&
          noOverlap = .true.

      end if
    end if

    if (noOverlap) then
      allocate(xoredRanges(1))
      xoredRanges(1) = copyRange(range)
  return
    end if

    ! If there is overlap, the number of xoredRanges can be 0, 1 or 2, depending on where the rangeSubset lies and the types of boundaries each range has at each end. Another way to think of it is that there can either be or not be a remnant at either end of range after rangeSubset has been removed. The subroutine has been written to check each end in turn and see if there is a remnant there.

    lowerEndRemnantExists = .false. ! default
    select case (range%lower%type)
    case(UNDEFINED)
      if(rangeSubset%lower%type .ne. UNDEFINED)&
        lowerEndRemnantExists = .true.

    case(INCLUSIVE)
      if  ((rangeSubset%lower%type == INCLUSIVE&
      .and. rangeSubset%lower%value >  range%lower%value)&
      .or. (rangeSubset%lower%type == EXCLUSIVE&
      .and. rangeSubset%lower%value >= range%lower%value))&
        lowerEndRemnantExists = .true.

    case(EXCLUSIVE)
      if   (rangeSubset%lower%type .ne. UNDEFINED&
      .and. rangeSubset%lower%value > range%lower%value)&
        lowerEndRemnantExists = .true.

    case default
call error('dss_aux_bug0',"Shouldn't ever happen, contact the code developer.")
    end select

    upperEndRemnantExists = .false. ! default
    select case (range%upper%type)
    case(UNDEFINED)
      if(rangeSubset%upper%type .ne. UNDEFINED)&
        upperEndRemnantExists = .true.

    case(INCLUSIVE)
      if  ((rangeSubset%upper%type == INCLUSIVE&
      .and. rangeSubset%upper%value <  range%upper%value)&
      .or. (rangeSubset%upper%type == EXCLUSIVE&
      .and. rangeSubset%upper%value <= range%upper%value))&
        upperEndRemnantExists = .true.

    case(EXCLUSIVE)
      if   (rangeSubset%upper%type .ne. UNDEFINED&
      .and. rangeSubset%upper%value < range%upper%value)&
        upperEndRemnantExists = .true.

    case default
call error('dss_aux_bug1',"Shouldn't ever happen, contact the code developer.")
    end select

    numFragments = 0
    if (lowerEndRemnantExists) numFragments = numFragments + 1
    if (upperEndRemnantExists) numFragments = numFragments + 1
    allocate(xoredRanges(numFragments))
 
    ! Now load up xoredRanges:

    fragmentNum = 0
    if (lowerEndRemnantExists) then
      fragmentNum = fragmentNum + 1
      xoredRanges(fragmentNum)%lower%type  = range%lower%type
      xoredRanges(fragmentNum)%lower%value = range%lower%value
      xoredRanges(fragmentNum)%upper%value = rangeSubset%lower%value

      if (rangeSubset%lower%type == INCLUSIVE) then
        xoredRanges(fragmentNum)%upper%value = EXCLUSIVE

      else ! rangeSubset%lower%type == EXCLUSIVE. UNDEFINED is ruled out
        ! in the test for lowerEndRemnantExists.

        xoredRanges(fragmentNum)%upper%value = INCLUSIVE
      end if
    end if
    
    if (upperEndRemnantExists) then
      fragmentNum = fragmentNum + 1
      xoredRanges(fragmentNum)%upper%type  = range%upper%type
      xoredRanges(fragmentNum)%upper%value = range%upper%value
      xoredRanges(fragmentNum)%lower%value = rangeSubset%upper%value

      if (rangeSubset%upper%type == INCLUSIVE) then
        xoredRanges(fragmentNum)%lower%value = EXCLUSIVE

      else ! rangeSubset%upper%type == EXCLUSIVE. UNDEFINED is ruled out
        ! in the test for upperEndRemnantExists.

        xoredRanges(fragmentNum)%lower%value = INCLUSIVE
      end if
    end if

  end subroutine xorRanges

  !---------------------------------------------------------------------
  subroutine correctRanges(ranges)
  !---------------------------------------------------------------------
!*** change to a function??

    implicit none

    type(RangeT), pointer :: ranges(:)

    type(RangeT) :: localRanges(size(ranges))
    integer :: numRanges, i, locVec(1), start, finis, rangeNum

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(ranges)
  if (numRanges .le. 0) return

    ! Swap any bounds which have the wrong internal order:
    do i = 1, numRanges
      if (ranges(i)%lower%type .ne. UNDEFINED&
      .and. ranges(i)%upper%type .ne. UNDEFINED&
      .and. ranges(i)%lower%value .gt. ranges(i)%upper%value) then
        call swapRangeBounds(ranges(i))
      end if
    end do

    ! Now sort them in order first of the lower bound, then the upper:
    call sortRanges(ranges)

    ! Find the last element of each array that has lo%type = UNDEFINED:
    if (any(ranges%lower%type .eq. UNDEFINED)) then
      locVec = maxloc(ranges%lower%value, (ranges%lower%type .eq. UNDEFINED))
      start = locVec(1)
    else
      start = 1
    end if

    ! Find the first element of each array that has hi%type = UNDEFINED:
    if (any(ranges%upper%type .eq. UNDEFINED)) then
      locVec = minloc(ranges%upper%value, (ranges%upper%type .eq. UNDEFINED))
      finis = locVec(1)
    else
      finis = 1
    end if

    ! Finally, transfer them to localRanges, amalgamating (ORing) overlapping
    ! members:
    rangeNum = 1
    localRanges(1) = ranges(start)
    do i = start + 1, finis
      if (checkRangeOverlap(localRanges(rangeNum), ranges(i)) .eq. BOVERLAPSA)&
      then
        localRanges(rangeNum) = orRanges(localRanges(rangeNum), ranges(i))
      else
        rangeNum = rangeNum + 1
        localRanges(rangeNum) = ranges(i)
      end if
    end do

    deallocate(ranges)
    allocate(ranges(rangeNum))
    ranges = localRanges

  end subroutine correctRanges

  !---------------------------------------------------------------------
  subroutine swapRangeBounds(range)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(inout) :: range

    type(BoundT) :: tempLo

    !-------------------------------------------------------------------
    ! Start

    tempLo      = range%lower
    range%lower = range%upper
    range%upper = tempLo

  end subroutine swapRangeBounds

  !---------------------------------------------------------------------
  subroutine sortRanges(ranges)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(inout) :: ranges(:)

    integer :: i, numRanges, start, finis

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(ranges)

    call sortBounds(ranges%lower, 'low')

    start = 1
    do while(start .le. numRanges)
      if (ranges(start)%lower%type .eq. UNDEFINED) then
        start = start + 1
    cycle
      end if

      finis = start
      do i = start + 1, numRanges
        if (ranges(i)%lower%value .ne. ranges(start)%lower%value&
        .or. ranges(i)%lower%type .ne. ranges(start)%lower%type) then
          finis = i - 1
        end if
      end do

      if (finis .gt. start) then
        call sortBounds(ranges(start:finis)%upper, 'hi')
      end if

      start = finis + 1
    end do

  end subroutine sortRanges

  !---------------------------------------------------------------------
  subroutine sortBounds(bounds, style)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(inout) :: bounds(:)
    character(*), intent(in) :: style

    integer :: index(size(bounds)), i, numBounds, locVec(1), start, finis
    type(BoundT) :: localBounds(size(bounds))

    !-------------------------------------------------------------------
    ! Start

    numBounds = size(bounds)

    if (style .eq. 'low') then
      ! First, move all the type = UNDEFINED down to the bottom:

      do i = 1, numBounds - 1
      if (bounds(i)%type .eq. UNDEFINED) cycle
      if (.not. any(bounds(i+1:)%type .eq. UNDEFINED)) exit

        locVec = minloc(bounds(i+1:)%value&
        , (bounds(i+1:)%type .eq. UNDEFINED))

        call swapBounds(bounds(i), bounds(locVec(1)))
      end do

      if (any(bounds%type .ne. UNDEFINED)) then ! sort them:
        do i = 1, numBounds
          if (bounds(i)%type .ne. UNDEFINED) then
            start = i
        exit
          end if
        end do

        call quickSortStatic(bounds(start:)%value, index(start:))
        do i = 1, start - 1
          localBounds(i) = bounds(i)
        end do
        do i = start, numBounds
          localBounds(i) = bounds(index(i))
        end do
      end if

    else ! style .eq. 'hi'
      ! First, move all the type = UNDEFINED up to the top:

      do i = numBounds, 2, -1
      if (bounds(i)%type .eq. UNDEFINED) cycle
      if (.not. any(bounds(:i-1)%type .eq. UNDEFINED)) exit

        locVec = minloc(bounds(:i-1)%value&
        , (bounds(:i-1)%type .eq. UNDEFINED))

        call swapBounds(bounds(i), bounds(locVec(1)))
      end do

      if (any(bounds%type .ne. UNDEFINED)) then ! sort them:
        do i = numBounds, 1, -1
          if (bounds(i)%type .ne. UNDEFINED) then
            finis = i
        exit
          end if
        end do

        call quickSortStatic(bounds(:finis)%value, index(:finis))
        do i = 1, finis
          localBounds(i) = bounds(index(i))
        end do
        do i = finis + 1, numBounds
          localBounds(i) = bounds(i)
        end do
      end if
    end if

  end subroutine sortBounds

  !---------------------------------------------------------------------
  subroutine swapBounds(boundA, boundB)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(inout) :: boundA, boundB

    type(BoundT) :: tempBound

    !-------------------------------------------------------------------
    ! Start

    tempBound = boundA
    boundA = boundB
    boundB = tempBound

  end subroutine swapBounds

  !---------------------------------------------------------------------
  subroutine quickSortStatic(values, index)
  !---------------------------------------------------------------------
!*** move it to sort_mod??

!    use sort_mod

    implicit none

    real(double), intent(in) :: values(:)
    integer,      intent(out) :: index(size(values))

    real(double) :: tempValues(size(values))

    !-------------------------------------------------------------------
    ! Start

    tempValues = values
    call quick_sort(tempValues, index)

  end subroutine quickSortStatic

  !---------------------------------------------------------------------
  function filterTypeToStr(filterType) result(filterTypeStr)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: filterType
    character(20) :: filterTypeStr

    !-------------------------------------------------------------------
    ! Start

    select case(filterType)
    case(RANGEFILTERT)
      filterTypeStr = 'RANGEFILTERT'
    case(REGIONFILTERT)
      filterTypeStr = 'REGIONFILTERT'
    case(BITMASKFILTERT)
      filterTypeStr = 'BITMASKFILTERT'
    case(RANGEBLOCKFILTERT)
      filterTypeStr = 'RANGEBLOCKFILTERT'
    case(MASKFILTERT)
      filterTypeStr = 'MASKFILTERT'
    case(REGIONBLOCKFILTERT)
      filterTypeStr = 'REGIONBLOCKFILTERT'
    case(GTIFILTERT)
      filterTypeStr = 'GTIFILTERT'
    case default
      filterTypeStr = 'UNKNOWN'
    end select

  end function filterTypeToStr

  !---------------------------------------------------------------------
  function rangeTypeToStr(rangeType) result(rangeTypeStr)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: rangeType
    character(9) :: rangeTypeStr

    !-------------------------------------------------------------------
    ! Start

    select case(rangeType)
    case(UNDEFINED)
      rangeTypeStr = 'UNDEFINED'
    case(INCLUSIVE)
      rangeTypeStr = 'INCLUSIVE'
    case(EXCLUSIVE)
      rangeTypeStr = 'EXCLUSIVE'
    case default
      rangeTypeStr = 'UNKNOWN  '
    end select

  end function rangeTypeToStr

  !---------------------------------------------------------------------
  integer function numDssCompsOfAxis(dssPointer, axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName

    integer :: compNum
    type(DScompT) :: dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssCompsOfAxis = 0
    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
      if (anyFilterOfThisAxis(dssComp, axisName))&
        numDssCompsOfAxis = numDssCompsOfAxis + 1
    end do

  end function numDssCompsOfAxis

  !---------------------------------------------------------------------
  integer function numDssCompsWithValue(dssPointer, axisName, axisValue)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    integer :: compNum
    type(DScompT) :: dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssCompsWithValue = 0 ! default

    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
      if (allFiltersPassValue(dssComp, axisName, axisValue))&
        numDssCompsWithValue = numDssCompsWithValue + 1
    end do

  end function numDssCompsWithValue

  !---------------------------------------------------------------------
  subroutine getNum1stDssCompWithValue(dssPointer, axisName, axisValue&
  , firstCompNum, status)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue
    integer, intent(out) :: firstCompNum, status

    integer :: compNum
    type(DScompT) :: dssComp
!    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    status = -1		! default to bad status value.
    firstCompNum = -1	! default to invalid value.

    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
      if (allFiltersPassValue(dssComp, axisName, axisValue)) then
        status = 0
        firstCompNum = compNum
    exit
      end if
    end do

  end subroutine getNum1stDssCompWithValue

  !---------------------------------------------------------------------
  logical function anyFilterOfThisAxis(dssComp, axisName)
  !---------------------------------------------------------------------
  ! This returns true only if the specified component contains at least one filter on the specified axis.

    implicit none

    type(DScompT),   intent(in) :: dssComp
    character(*), intent(in) :: axisName

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    numFilters = dssFilters(dssComp)
    anyFilterOfThisAxis = .false.	! default

    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .eq. axisName) then
        anyFilterOfThisAxis = .true.
  return
      end if
    end do

  end function anyFilterOfThisAxis

  !---------------------------------------------------------------------
  integer function numFiltersOfAxis(dssComp, axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT), intent(in) :: dssComp
    character(*),  intent(in) :: axisName
!    real(single),  intent(in) :: axisValue

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    numFiltersOfAxis = 0  ! default

    numFilters = dssFilters(dssComp)
    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .eq. axisName)&
        numFiltersOfAxis = numFiltersOfAxis + 1 
    end do

  end function numFiltersOfAxis

  !---------------------------------------------------------------------
  subroutine get1stFilterThisAxis(dssComp, axisName, firstFilter, status)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT),   intent(in)  :: dssComp
    character(*),    intent(in)  :: axisName
    type(DSfilterT), intent(out) :: firstFilter
    integer,         intent(out) :: status

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    status = -1		! default to bad status value.

    numFilters = dssFilters(dssComp)
    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .eq. axisName) then
        status = 0
        firstFilter = filter
    exit
      end if
    end do

  end subroutine get1stFilterThisAxis

  !---------------------------------------------------------------------
  logical function allFiltersPassValue(dssComp, axisName, axisValue)
  !---------------------------------------------------------------------
  ! This returns true ONLY if the specified component contains at least one filter on the specified axis AND all of the filters of this component and on this axis pass the specified value.

    implicit none

    type(DScompT),   intent(in) :: dssComp
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical, allocatable :: filterOnThisAxisExists(:), valueIsPassed(:)

    !-------------------------------------------------------------------
    ! Start

    numFilters = dssFilters(dssComp)
    allocate(filterOnThisAxisExists(0:numFilters - 1)&
    , valueIsPassed(0:numFilters - 1))
    filterOnThisAxisExists = .false.	! default
    valueIsPassed          = .false.	! default

    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      if (dssFilterName(filter) .eq. axisName) then
        filterOnThisAxisExists(filtNum) = .true.
        if (valueIsPassedByDssFilter(axisValue, filter))&
          valueIsPassed(filtNum) = .true. 
      end if
    end do

    if (.not. any(filterOnThisAxisExists) .or. any(filterOnThisAxisExists&
    .and. .not. valueIsPassed)) then
      allFiltersPassValue = .false.
    else
      allFiltersPassValue = .true.
    end if

  end function allFiltersPassValue

  !---------------------------------------------------------------------
  logical function valueIsPassedByComp(dssComp, axisName, axisValue)
  !---------------------------------------------------------------------
  ! This returns true ONLY if the specified component contains no filter on the specified axis OR the value is passed by all of the filters of this component that do exist for the given axis.

    implicit none

    type(DScompT),intent(in) :: dssComp
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical, allocatable :: filterIsOnSpecifiedAxis(:), valueIsPassed(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'start of valueIsPassedByComp'
if (testFlag) print *, 'axisName: ', axisName
    numFilters = dssFilters(dssComp)
if (testFlag) print *, 'numFilters in comp: ', numFilters
    allocate(filterIsOnSpecifiedAxis(numFilters), valueIsPassed(numFilters))
    filterIsOnSpecifiedAxis = .false.	! default
    valueIsPassed           = .false.	! default

    do filtNum = 1, numFilters
if (testFlag) print *, '  filtNum: ', filtNum
      filter = dssFilter(dssComp, filtNum - 1)
if (testFlag) print *, '  dssFilterName(filter): ', dssFilterName(filter)
      if (dssFilterName(filter) .eq. axisName) then
if (testFlag) print *, '  filterIsOnSpecifiedAxis(filtNum) = .true.'
        filterIsOnSpecifiedAxis(filtNum) = .true.
        if (valueIsPassedByDssFilter(axisValue, filter)) then
if (testFlag) print *, '  valueIsPassed(',filtNum,') = .true.'
          valueIsPassed(filtNum) = .true.
else
if (testFlag) print *, '  valueIsPassed(',filtNum,') = .false.'
end if
      end if
    end do
if (testFlag) print *, '  filtNum: ', filtNum

    if (all(filterIsOnSpecifiedAxis .eqv. valueIsPassed)) then
      valueIsPassedByComp = .true.
    else
      valueIsPassedByComp = .false.
    end if
if (testFlag) print *, 'end of valueIsPassedByComp'

  end function valueIsPassedByComp

  !---------------------------------------------------------------------
  logical function valueIsPassedByDssFilter(value, filter)
  !---------------------------------------------------------------------

!    use fits_utils ! for readColData.

    implicit none

    real(single), intent(in) :: value
    type(DSfilterT), intent(in) :: filter

    type(RangeT), pointer :: ranges(:)
    character(255)  :: gtiBlockName
    type(TableT)    :: tab
    integer         :: numGtis, i
    type(RangeT), allocatable :: gti(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'start of valueIsPassedByDssFilter'
    valueIsPassedByDssFilter = .false.

    select case(dssFilterType(filter))
    case(RANGEFILTERT)
if (testFlag) print *, 'dssFilterType(filter) = RANGEFILTERT'
      ranges => dssRanges(filter)
if (testFlag) print *, 'size(ranges): ', size(ranges)
if (testFlag) print *, 'value: ', value
      do i = 0, size(ranges) - 1
if (testFlag) print *, '  i: ', i
        if (valueIsWithinRange(value, ranges(i))) then
          valueIsPassedByDssFilter = .true.
      exit
        end if
      end do

deallocate(ranges)

!    case(REGIONFILTERT)
!      selectionExpression = selectExpression(filter)
!*** for a future version of valueIsPassedByDssFilter() with 2-value
!*** alternative input: set up pseudo-table with 1 entry, filter it,
!*** count num filtered rows.
!    case(BITMASKFILTERT)
!    case(RANGEBLOCKFILTERT)
!    case(MASKFILTERT)
!    case(REGIONBLOCKFILTERT)

    case(GTIFILTERT) ! Treat GTIs as ranges:
if (testFlag) print *, 'dssFilterType(filter) = GTIFILTERT'
      gtiBlockName = dssFilterBlock(filter)
      tab = table(dataSet(gtiBlockName, READ), gtiBlockName)
      numGtis = numberOfRows(tab)
      allocate(gti(numGtis))
      call readColData(tab, 'START', gti%lower%value)
      call readColData(tab, 'STOP',  gti%upper%value)
      call release(tab)

      gti%lower%type = INCLUSIVE
      gti%upper%type = INCLUSIVE
      do i = 1, size(gti)
        if (valueIsWithinRange(value, gti(i))) then
          valueIsPassedByDssFilter = .true.
      exit
        end if
      end do

    case default
call error('','')
    end select

if (testFlag) print *, 'end of valueIsPassedByDssFilter'
  end function valueIsPassedByDssFilter

  !---------------------------------------------------------------------
  logical function valueIsWithinRange(value, range)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: value
    type(RangeT), intent(in) :: range

    !-------------------------------------------------------------------
    ! Start

    valueIsWithinRange = .false. ! default

!print *, '    value: ', value
!print *, '    range%lower%type: ', rangeTypeToStr(range%lower%type)
!if   (range%lower%type .ne. UNDEFINED) then
!  print *, '    range%lower%value: ', range%lower%value
!end if
!print *, '    range%upper%type: ', rangeTypeToStr(range%upper%type)
!if   (range%upper%type .ne. UNDEFINED) then
!  print *, '    range%upper%value: ', range%upper%value
!end if

    if   (range%lower%type .eq. UNDEFINED&
    .or. (range%lower%type .eq. INCLUSIVE .and. range%lower%value <= value)&
    .or. (range%lower%type .eq. EXCLUSIVE .and. range%lower%value < value))&
    then
      if   (range%upper%type .eq. UNDEFINED&
      .or. (range%upper%type .eq. INCLUSIVE .and. range%upper%value >= value)&
      .or. (range%upper%type .eq. EXCLUSIVE .and. range%upper%value > value))&
      then
        valueIsWithinRange = .true.
      end if
    end if

  end function valueIsWithinRange

  !---------------------------------------------------------------------
  subroutine dumpDssBlock(inBlock)
  !---------------------------------------------------------------------

    implicit none

    type(BlockT), intent(in) ::&
      inBlock

    type(DssT) ::&
      dssPointer

    !-------------------------------------------------------------------
    ! Start:

    dssPointer = dataSubspace(inBlock)
    call dumpDssPointer(dssPointer)

  end subroutine dumpDssBlock

  !---------------------------------------------------------------------
  subroutine dumpDssPointer(dssPointer)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) ::&
      dssPointer

    type(DScompT) :: dssComp
    integer :: numComps, compNum

    !-------------------------------------------------------------------
    ! Start:

    numComps = dssComponents(dssPointer)
    if (numComps <= 0) then
      print *, 'DSS contains no components.'
  return
    end if

    do compNum = 0, numComps - 1
      print *, 'DSS component number', compNum
      dssComp = dssComponent(dssPointer, compNum)
      call dumpDssComponent(dssComp)
    end do

  end subroutine dumpDssPointer

  !---------------------------------------------------------------------
  subroutine dumpDssComponent(dssComp)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT), intent(in) ::&
      dssComp

    type(DSfilterT) ::&
      filter
    type(RangeT), dimension(:), pointer :: range
    character(255) ::&
      blockName
    integer(kind=int16) ::&
      i
    integer(kind=int32) ::&
      bitmask
    integer :: filtNum, filterType
    logical :: typeNotFound
    character(len=20) :: columnName, filterTypeStr
    character(len=80) :: axisNames
    character(8) :: bitMaskStr
    type filtInfo
      integer :: index
      character(len=20) :: name
    end type
    type(filtInfo) :: myFiltInfo(7)

    !-------------------------------------------------------------------
    ! Start:

    myFiltInfo(1)%index = RANGEFILTERT
    myFiltInfo(2)%index = REGIONFILTERT
    myFiltInfo(3)%index = BITMASKFILTERT
    myFiltInfo(4)%index = RANGEBLOCKFILTERT
    myFiltInfo(5)%index = GTIFILTERT
    myFiltInfo(6)%index = MASKFILTERT
    myFiltInfo(7)%index = REGIONBLOCKFILTERT
    myFiltInfo(1)%name = 'RANGEFILTERT'
    myFiltInfo(2)%name = 'REGIONFILTERT'
    myFiltInfo(3)%name = 'BITMASKFILTERT'
    myFiltInfo(4)%name = 'RANGEBLOCKFILTERT'
    myFiltInfo(5)%name = 'GTIFILTERT'
    myFiltInfo(6)%name = 'MASKFILTERT'
    myFiltInfo(7)%name = 'REGIONBLOCKFILTERT'

    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
      columnName = dssFilterName(filter)
      filterType = dssFilterType(filter)

      typeNotFound = .true.
      do i = 1, 7
        if (myFiltInfo(i)%index .eq. filterType) then
          typeNotFound = .false.
          filterTypeStr = myFiltInfo(i)%name
      exit
        end if
      end do
!print *, 'filterTypeStr', filterTypeStr
if (typeNotFound) call error('unknownFilterType','')

      write(*, '(a, i3.3, a)') '    Filter=', filtNum&
        , ', axis='//trim(columnName)//', type='//filterTypeStr

!print *, 'filterType: ',filterType
!print *, BITMASKFILTERT

      select case(filterType)
      case(RANGEFILTERT)
        range => dssRanges(filter)
        call dumpRangeChain(range)

      case(REGIONBLOCKFILTERT)
        blockName = dssFilterBlock(filter)
        print *, '        Block name = ', trim(blockName)

      case(BITMASKFILTERT)
        bitmask = dssFilterMask(filter)
!print *, 'aaa'
!print *, 'bitmask: ', bitmask
!print *, 'bbb'
        write(bitMaskStr, '(z8)') bitmask
!print *, 'ccc'
!print *, 'bitMaskStr: ', bitMaskStr
!print *, 'ddd'
!        print *, '        Bitmask = 0x', trim(adjustl(bitMaskStr))
        print *, '        Bitmask = 0x', bitMaskStr
!print *, 'eee'

      case(RANGEBLOCKFILTERT)
        blockName = dssFilterBlock(filter)
        print *, '        Block name = ', trim(blockName)

      case(GTIFILTERT)
        blockName = dssFilterBlock(filter)
        print *, '        Block name = ', trim(blockName)

      case(MASKFILTERT, REGIONFILTERT)
        axisNames = dssFilterComps(filter)
        print *, '        Axis names', axisNames
      end select
    end do
    print *, ''

  end subroutine dumpDssComponent

  !---------------------------------------------------------------------
  subroutine dumpDssPointer_old(dssPointer)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) ::&
      dssPointer

    type(DSfilterT) ::&
      filter
    type(RangeT), dimension(:), pointer :: range
    character(255) ::&
      blockName
    integer(kind=int16) ::&
      i
    integer(kind=int32) ::&
      bitmask
    integer :: compNum, filtNum, filterType
    logical :: typeNotFound
    character(len=20) :: columnName, filterTypeStr
!, dssKwdValue
    character(len=80) :: axisNames
    character(8) :: bitMaskStr
    type filtInfo
      integer :: index
      character(len=20) :: name
    end type
    type(filtInfo) :: myFiltInfo(7)

    !-------------------------------------------------------------------
    ! Start:

    myFiltInfo(1)%index = RANGEFILTERT
    myFiltInfo(2)%index = REGIONFILTERT
    myFiltInfo(3)%index = BITMASKFILTERT
    myFiltInfo(4)%index = RANGEBLOCKFILTERT
    myFiltInfo(5)%index = GTIFILTERT
    myFiltInfo(6)%index = MASKFILTERT
    myFiltInfo(7)%index = REGIONBLOCKFILTERT
    myFiltInfo(1)%name = 'RANGEFILTERT'
    myFiltInfo(2)%name = 'REGIONFILTERT'
    myFiltInfo(3)%name = 'BITMASKFILTERT'
    myFiltInfo(4)%name = 'RANGEBLOCKFILTERT'
    myFiltInfo(5)%name = 'GTIFILTERT'
    myFiltInfo(6)%name = 'MASKFILTERT'
    myFiltInfo(7)%name = 'REGIONBLOCKFILTERT'

    do compNum = 0, dssComponents(dssPointer) - 1
      print *, 'DSS component number', compNum
      do filtNum = 0, dssFilters(dssComponent(dssPointer, compNum)) - 1
        filter = dssFilter(dssComponent(dssPointer, compNum), filtNum)
        columnName = dssFilterName(filter)
        filterType = dssFilterType(filter)

        typeNotFound = .true.
        do i = 1, 7
          if (myFiltInfo(i)%index .eq. filterType) then
            typeNotFound = .false.
            filterTypeStr = myFiltInfo(i)%name
        exit
          end if
        end do
print *, 'filterTypeStr', filterTypeStr
if (typeNotFound) call error('unknownFilterType','')

!        print *, '    Filter=', filtNum, 'axis='//trim(columnName)&
!          //' type=', filterTypeStr
write(*, '(a, i3.3, a)') '    Filter=', filtNum&
, ', axis='//trim(columnName)//', type='//filterTypeStr

        select case(filterType)
        case(RANGEFILTERT)
          range => dssRanges(filter)
call dumpRangeChain(range)
!          do i = 0, size(range) - 1
!            print *, '        range number =', i
!
!            select case(range(i)%lower%type)
!            case(UNDEFINED)
!              print *, '          Lower bound = UNDEFINED.'
!            case(INCLUSIVE)
!              print *, '          Lower bound = ', range(i)%lower%value&
!                , ' (INCLUSIVE)'
!            case(EXCLUSIVE)
!              print *, '          Lower bound = ', range(i)%lower%value&
!                , ' (EXCLUSIVE)'
!            case default
!              print *, '          Lower bound = ', range(i)%lower%value&
!                , ' (type=???)'
!            end select
!
!            select case(range(i)%upper%type)
!            case(UNDEFINED)
!              print *, '          Upper bound = UNDEFINED.'
!            case(INCLUSIVE)
!              print *, '          Upper bound = ', range(i)%upper%value&
!                , ' (INCLUSIVE)'
!            case(EXCLUSIVE)
!              print *, '          Upper bound = ', range(i)%upper%value&
!                , ' (EXCLUSIVE)'
!            case default
!              print *, '          Upper bound = ', range(i)%upper%value&
!                , ' (type=???)'
!            end select
!          end do

        case(REGIONBLOCKFILTERT)
          blockName = dssFilterBlock(filter)
          print *, '        Block name = ', trim(blockName)

        case(BITMASKFILTERT)
          bitmask = dssFilterMask(filter)
          write(bitMaskStr, '(z8)') bitmask
          print *, '        Bitmask = 0x', trim(adjustl(bitMaskStr))

        case(RANGEBLOCKFILTERT)
          blockName = dssFilterBlock(filter)
          print *, '        Block name = ', trim(blockName)

        case(GTIFILTERT)
          blockName = dssFilterBlock(filter)
          print *, '        Block name = ', trim(blockName)

        case(MASKFILTERT, REGIONFILTERT)
          axisNames = dssFilterComps(filter)
          print *, '        Axis names', axisNames
        end select
      end do
      print *, ''
    end do

  end subroutine dumpDssPointer_old

  !---------------------------------------------------------------------
  logical function filterSetIsIdentical(dssCompA, dssCompB, axisName)
  !---------------------------------------------------------------------
  ! This returns true only if either (i) both components have no filter of the given axis name, or (ii) both components have the same number, type and values of filters of the given axis name.
  !
  ! Probably would be good to either extend this or clone a variant that compared ranges between filters of different but partially compatible types (range, rangeBlock, gti) and returned true if the ranges were identical, regardless of differences in the type.

    implicit none

    type(DScompT), intent(in) :: dssCompA, dssCompB
    character(*),  intent(in) :: axisName

    integer :: numFiltersA, numFiltersB, filtNumA, filtNumB, filtNumA1, filtNumA2, filtNumB1, filtNumB2
    logical, allocatable :: filtersThisAxisA(:), filtersThisAxisB(:)
    type(DSfilterT) :: filterA1, filterA2, filterB1, filterB2, filterA, filterB
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    filterSetIsIdentical: axis name is '//trim(axisName)//'.'

    filterSetIsIdentical = .false. ! default

    if   (.not. anyFilterOfThisAxis(dssCompA, axisName)&
    .and. .not. anyFilterOfThisAxis(dssCompB, axisName)) then
      filterSetIsIdentical = .true.
if (testFlag) print *, '    filterSetIsIdentical:   no filter this name in either comp.'
  return
    end if

    numFiltersA = dssFilters(dssCompA)
    allocate(filtersThisAxisA(numFiltersA))
    filtersThisAxisA = .false. ! default
    do filtNumA = 1, numFiltersA
      if (dssFilterName(dssFilter(dssCompA, filtNumA - 1)) == axisName) then
        filtersThisAxisA(filtNumA) = .true.
if (testFlag) print *, '    filterSetIsIdentical:   compA has a filter of this name, filter number', filtNumA
end if
    end do

    ! Knock out duplicated dssCompA filters:

    do filtNumA1 = 1, numFiltersA - 1
    if (.not. filtersThisAxisA(filtNumA1)) cycle
      filterA1 = dssFilter(dssCompA, filtNumA1 - 1)
      do filtNumA2 = filtNumA1 + 1, numFiltersA
      if (.not. filtersThisAxisA(filtNumA2)) cycle
        filterA2 = dssFilter(dssCompA, filtNumA2 - 1)
        if (filtersAreIdentical(filterA1, filterA2))&
          filtersThisAxisA(filtNumA2) = .false.
      end do
    end do

    numFiltersB = dssFilters(dssCompB)
    allocate(filtersThisAxisB(numFiltersB))
    filtersThisAxisB = .false. ! default
    do filtNumB = 1, numFiltersB
      if (dssFilterName(dssFilter(dssCompB, filtNumB - 1)) == axisName) then
        filtersThisAxisB(filtNumB) = .true.
if (testFlag) print *, '    filterSetIsIdentical:   compB has a filter of this name, filter number', filtNumB
end if
    end do

    ! Knock out duplicated dssCompB filters:

    do filtNumB1 = 1, numFiltersB - 1
    if (.not. filtersThisAxisB(filtNumB1)) cycle
      filterB1 = dssFilter(dssCompB, filtNumB1 - 1)
      do filtNumB2 = filtNumB1 + 1, numFiltersB
      if (.not. filtersThisAxisB(filtNumB2)) cycle
        filterB2 = dssFilter(dssCompB, filtNumB2 - 1)
        if (filtersAreIdentical(filterB1, filterB2))&
          filtersThisAxisB(filtNumB2) = .false.
      end do
    end do

  if (count(filtersThisAxisA) .ne. count(filtersThisAxisB)) return

    ! Now see if filters from A and B all match up. Each match-up is marked by setting the appropriate element of filtersThisAxisA (an arbitary choice, could equally well have chosen B for the honour) to false. Thus if any elements of filtersThisAxisA remain true at the end of the looping, this means not every filter was matched.

    ALOOP: do filtNumA = 1, numFiltersA
    if (.not. filtersThisAxisA(filtNumA)) cycle
      filterA = dssFilter(dssCompA, filtNumA - 1)

      BLOOP: do filtNumB = 1, numFiltersB
      if (.not. filtersThisAxisB(filtNumB)) cycle
        filterB = dssFilter(dssCompB, filtNumB - 1)

        if (filtersAreIdentical(filterA, filterB)) then
          filtersThisAxisA(filtNumA) = .false.
if (testFlag) print *, '    filterSetIsIdentical:   for filtNumA=',filtNumA&
,' and filtNumB=',filtNumB,', filters are identical. Exit Bloop.'
      exit BLOOP
        end if

      end do BLOOP
    end do ALOOP

    if (.not. any(filtersThisAxisA)) filterSetIsIdentical = .true.

if (testFlag) print *, '    filterSetIsIdentical:   filterSetIsIdentical'

  end function filterSetIsIdentical

  !---------------------------------------------------------------------
  logical function filtersAreIdentical(filterA, filterB)
  !---------------------------------------------------------------------

    implicit none

    type(DSfilterT), intent(in) :: filterA, filterB

    type(RangeT), pointer :: rangesA(:), rangesB(:)
    integer :: i, numRowsA, numRowsB, row
    character(255) :: blockNameA, blockNameB
    character(20) :: xyNameString, xColName, yColName, shapeA, shapeB
    type(TableT) :: tableA, tableB
    real(double), pointer :: colPtrA(:,:), colPtrB(:,:)
    type(ColumnT) :: colA, colB
    type(ArrayT) :: arrayA, arrayB
    integer(int32), pointer :: imageA(:,:), imageB(:,:)

    !-------------------------------------------------------------------
    ! Start

    filtersAreIdentical = .false. ! default

  if (dssFilterName(filterA)  .ne. dssFilterName(filterB))  return
  if (dssFilterType(filterA)  .ne. dssFilterType(filterB))  return
  if (dssFilterUnits(filterA) .ne. dssFilterUnits(filterB)) return

    select case(dssFilterType(filterA))
    case(RANGEFILTERT)
!      rangesA => oldToNewPtr(dssRanges(filterA))
!      rangesB => oldToNewPtr(dssRanges(filterB))
rangesA => dssRanges(filterA)
rangesB => dssRanges(filterB)

  if (size(rangesA) .ne. size(rangesB)) return

      do i = 0, size(rangesA) - 1
        if  (rangesA(i)%lower%type  .ne. rangesB(i)%lower%type&
        .or. rangesA(i)%lower%value .ne. rangesB(i)%lower%value&
        .or. rangesA(i)%upper%type  .ne. rangesB(i)%upper%type&
        .or. rangesA(i)%upper%value .ne. rangesB(i)%upper%value)&
  return
      end do

    case(REGIONBLOCKFILTERT)
      blockNameA = dssFilterBlock(filterA)
      blockNameB = dssFilterBlock(filterB)
      tableA = table(dataSet(blockNameA, READ), blockNameA)
      tableB = table(dataSet(blockNameB, READ), blockNameB)
      if (.not. columnSetIsIdentical(tableA, tableB)) then ! in fits_utils
        call release(tableA)
        call release(tableB)
  return
      end if

      numRowsA = numberOfRows(tableA)
      numRowsB = numberOfRows(tableB)
      if (numRowsA .ne. numRowsB) then
        call release(tableA)
        call release(tableB)
  return
      end if

      xyNameString = stringAttribute(tableA, 'MFORM1')
      call decomposeXyNameString(xyNameString, xColName, yColName)

      call readColData(tableA, xColName, colPtrA)
      call readColData(tableB, xColName, colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
  return
      end if
!*** possibly could be fooled by nulls

      call readColData(tableA, yColName, colPtrA)
      call readColData(tableB, yColName, colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

      if(hasColumn(tableA, 'SHAPE')) then
        colA = column(tableA, 'SHAPE')
        colB = column(tableB, 'SHAPE')
        do row = 0, numRowsA - 1
          shapeA = ucase(stringCell(colA, row))
          shapeB = ucase(stringCell(colB, row))
          if (shapeA .ne. shapeB) then
            call release(tableA)
            call release(tableB)
  return
          end if
        end do
      end if

      if(hasColumn(tableA, 'R')) then
        call readColData(tableA, 'R', colPtrA)
        call readColData(tableB, 'R', colPtrB)
        if (any(colPtrA .ne. colPtrB)) then
          call release(tableA)
          call release(tableB)
  return
        end if
!*** possibly could be fooled by nulls
      end if

      if(hasColumn(tableA, 'ROTANG')) then
        call readColData(tableA, 'ROTANG', colPtrA)
        call readColData(tableB, 'ROTANG', colPtrB)
        if (any(colPtrA .ne. colPtrB)) then
          call release(tableA)
          call release(tableB)
  return
        end if
!*** possibly could be fooled by nulls
      end if

    case(BITMASKFILTERT)
  if(dssFilterMask(filterA) .ne. dssFilterMask(filterB)) return

    case(GTIFILTERT)
      blockNameA = dssFilterBlock(filterA)
      blockNameB = dssFilterBlock(filterB)
      tableA = table(dataSet(blockNameA, READ), blockNameA)
      tableB = table(dataSet(blockNameB, READ), blockNameB)

      numRowsA = numberOfRows(tableA)
      numRowsB = numberOfRows(tableB)
      if (numRowsA .ne. numRowsB) then
        call release(tableA)
        call release(tableB)
  return
      end if

      call readColData(tableA, 'TSTART', colPtrA)
      call readColData(tableB, 'TSTART', colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
  return
      end if
!*** possibly could be fooled by nulls

      call readColData(tableA, 'TSTOP', colPtrA)
      call readColData(tableB, 'TSTOP', colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableA)
        call release(tableB)
  return
      end if
!*** possibly could be fooled by nulls

    case(MASKFILTERT)
      blockNameA = dssFilterBlock(filterA)
      blockNameB = dssFilterBlock(filterB)
      arrayA = array(dataSet(blockNameA, READ), blockNameA, READ)
      arrayB = array(dataSet(blockNameB, READ), blockNameB, READ)
      call loadFitsArray(arrayA, imageA)
      call loadFitsArray(arrayB, imageB)

      if  (size(imageA, 1) .ne. size(imageB, 1)&
      .or. size(imageA, 2) .ne. size(imageB, 2)) then
        deallocate(imageA, imageB)
        call release(arrayA)
        call release(arrayB)
  return
      end if

      if  (any(imageA .ne. imageB)) then
        deallocate(imageA, imageB)
        call release(arrayA)
        call release(arrayB)
  return
      end if

!    case(RANGEBLOCKFILTERT)	! doesn't seem to be any firm convention
!				! for it yet.

!    case(REGIONFILTERT)	! not normally used in the DSS.

    case default
call error('', 'ssclib: ')
    end select

    filtersAreIdentical = .true. ! Made it through all the tests!

  end function filtersAreIdentical

  !---------------------------------------------------------------------
  subroutine decomposeXyNameString(xyNameString, xColName, yColName)
  !---------------------------------------------------------------------
  ! Designed to analyse the output of the call dssFilterComps, which returns a string containing the names of the 'X' and 'Y' columns of a region specification block separated by a comma.

    implicit none

    character(*), intent(in) :: xyNameString
    character(*), intent(out) :: xColName, yColName

    integer :: posOfComma

    !-------------------------------------------------------------------
    ! Start

    xColName = '' ! default
    yColName = '' ! default

    posOfComma = index(xyNameString, ',')
if (posOfComma == 0) call error('commaNotFound', 'ssclib:')

    xColName = trim(adjustl(xyNameString(:posOfComma-1)))
    yColName = trim(adjustl(xyNameString(posOfComma+1:)))

  end subroutine decomposeXyNameString

  !---------------------------------------------------------------------
  function getRangeSpan(dssPointer, axisName) result(dssRangeSpan)
  !---------------------------------------------------------------------
  ! This subroutine does a logical OR of all the ranges of all the
  ! components in the DSS and returns the extrema of that ORed range.

    implicit none

    type(DssT), intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    type(RangeT) :: dssRangeSpan

    type(DScompT) :: dssComp
    type(RangeT), pointer  :: oredRanges(:), compRanges(:), tempOredRanges(:)
    integer(int32) :: numComps, compNum, numRanges

    !-------------------------------------------------------------------
    ! Start

!print *, '    >>> entered getRangeSpan'
    numComps = dssComponents(dssPointer)
!print *, '    axisName: ', axisName
!print *, '    numComps: ', numComps
    if (numComps .le. 0) then
      dssRangeSpan%lower%type = UNDEFINED
      dssRangeSpan%lower%value = 0.0
      dssRangeSpan%upper%type = UNDEFINED
      dssRangeSpan%upper%value = 0.0
  return
    end if

    dssComp = dssComponent(dssPointer, 0)
    call loadRangesComp(dssComp, axisName, oredRanges)
!print *, '    size(ranges) from 0th comp: ', size(oredRanges)

    do compNum = 1, numComps - 1
!print *, '      i: ', compNum
      dssComp = dssComponent(dssPointer, compNum)
      call loadRangesComp(dssComp, axisName, compRanges)
!print *, '      size(ranges) from ith comp: ', size(compRanges)
      tempOredRanges => orAllRanges(compRanges, oredRanges)
      deallocate(oredRanges) ! clears the memory
      oredRanges => tempOredRanges
    end do
!print *, '    size(oredRanges) at end of loop: ', size(oredRanges)

    numRanges = size(oredRanges)
    dssRangeSpan%lower%type  = oredRanges(1)%lower%type
    dssRangeSpan%lower%value = oredRanges(1)%lower%value
    dssRangeSpan%upper%type  = oredRanges(numRanges)%upper%type
    dssRangeSpan%upper%value = oredRanges(numRanges)%upper%value

    deallocate(oredRanges) ! also clears tempOredRanges, since they
    ! are associated and thus point to the same memory.

!print *, '    <<< leaving getRangeSpan'

  end function getRangeSpan

  !---------------------------------------------------------------------
  subroutine dumpRangeChain(rangeChain)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeChain(:)

    integer(int32) :: rangeNum

    !-------------------------------------------------------------------
    ! Start

    do rangeNum = 1, size(rangeChain)
      print *, '        range number =', rangeNum

      select case(rangeChain(rangeNum)%lower%type)
      case(UNDEFINED)
        print *, '          Lower bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '          Lower bound = ', rangeChain(rangeNum)%lower%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '          Lower bound = ', rangeChain(rangeNum)%lower%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '          Lower bound = ', rangeChain(rangeNum)%lower%value&
          , ' (type=???)'
      end select

      select case(rangeChain(rangeNum)%upper%type)
      case(UNDEFINED)
        print *, '          Upper bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '          Upper bound = ', rangeChain(rangeNum)%upper%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '          Upper bound = ', rangeChain(rangeNum)%upper%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '          Upper bound = ', rangeChain(rangeNum)%upper%value&
          , ' (type=???)'
      end select
    end do

  end subroutine dumpRangeChain

  !---------------------------------------------------------------------
  subroutine dumpBoxChain(boxChain)
  !---------------------------------------------------------------------
!  type, public :: rangeBoxType
!    type(RangeT) :: x, y
!  end type rangeBoxType
!  type, public :: boxChainType
!    type(rangeBoxType), pointer :: boxes(:)
!    integer(int32), pointer :: comps(:)
!    character(20) :: xAxisName, yAxisName
!  end type boxChainType

    implicit none

    type(boxChainType), intent(in) :: boxChain

    integer(int32) :: boxNum, compNum

    !-------------------------------------------------------------------
    ! Start

    do boxNum = 1, size(boxChain%boxes)
      print *, 'range box number =', boxNum

      print *, '  X axis name is ', trim(boxChain%xAxisName)
      select case(boxChain%boxes(boxNum)%x%lower%type)
      case(UNDEFINED)
        print *, '    Lower bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%x%lower%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%x%lower%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%x%lower%value&
          , ' (type=???)'
      end select

      select case(boxChain%boxes(boxNum)%x%upper%type)
      case(UNDEFINED)
        print *, '    Upper bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%x%upper%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%x%upper%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%x%upper%value&
          , ' (type=???)'
      end select

      print *, '  Y axis name is ', trim(boxChain%yAxisName)
      select case(boxChain%boxes(boxNum)%y%lower%type)
      case(UNDEFINED)
        print *, '    Lower bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%y%lower%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%y%lower%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%y%lower%value&
          , ' (type=???)'
      end select

      select case(boxChain%boxes(boxNum)%y%upper%type)
      case(UNDEFINED)
        print *, '    Upper bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%y%upper%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%y%upper%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%y%upper%value&
          , ' (type=???)'
      end select
    end do

    print *, 'Valid DSS component numbers are:'
    do compNum = 1, size(boxChain%comps)
      print *, '  ', boxChain%comps(compNum)
    end do

  end subroutine dumpBoxChain

  !---------------------------------------------------------------------
  logical function valuesArePassedByBoxChain(xValue, yValue, boxChain)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: xValue, yValue
    type(boxChainType), intent(in) :: boxChain

    integer(int32) :: boxNum, numBoxes

    !-------------------------------------------------------------------
    ! Start

    valuesArePassedByBoxChain = .false. ! default

    numBoxes = size(boxChain%boxes)
    do boxNum = 1, numBoxes
!print *, '  boxNum, xValue, yValue: ', boxNum, xValue, yValue
      if   (valueIsWithinRange(xValue, boxChain%boxes(boxNum)%x)&
      .and. valueIsWithinRange(yValue, boxChain%boxes(boxNum)%y)) then
        valuesArePassedByBoxChain = .true.
  return
      end if
!call error('valuesArePassedByBoxChain','Under test - stopping here.')
    end do

  end function valuesArePassedByBoxChain

  !---------------------------------------------------------------------
  function projectBoxChainToRangeChain(boxChain, projectToX) result(rangeChain)
  !---------------------------------------------------------------------
  ! Remember to deallocate rangeChain after use.

    implicit none

    type(boxChainType), intent(in) :: boxChain
    logical, intent(in) :: projectToX
    type(RangeT), pointer :: rangeChain(:)

    integer(int32) :: boxNum, numBoxes

    !-------------------------------------------------------------------
    ! Start

    numBoxes = size(boxChain%boxes)
    allocate(rangeChain(numBoxes))
    if (projectToX) then
      do boxNum = 1, numBoxes
        rangeChain(boxNum) = copyRange(boxChain%boxes(boxNum)%x)
      end do

    else ! projectToY:
      do boxNum = 1, numBoxes
        rangeChain(boxNum) = copyRange(boxChain%boxes(boxNum)%y)
      end do
    end if

    ! At this point, some of the ranges might overlap, and they are
    ! not necessarily in order. They have to be cleaned up.

    call correctRanges(rangeChain)

    ! Now they are ok.

  end function projectBoxChainToRangeChain

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dss_aux
