!
! Module: simulations
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to generate a simulated XMM EPIC image and to compare the list of simulated sources used to make that image with the list of sources detected in it.
! 
! Comments: TODO:
! 
! $Id: simulations.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module simulations

  use dal
  use cal
  use errorHandling

  use array_utils	! for rebin()
  use ssc_misc		! for indicize(), fracVal()
  use constants

  implicit none 

  private 

  public getCalPsfPatch
  public resizePatch

contains 

  !---------------------------------------------------------------------
  subroutine getCalPsfPatch(srcXpixels, srcYpixels, srcThetaArcsec&
  , fovCentreXpixels, fovCentreYpixels, energyFraction, EeV&
  , pixelSizeXmm, pixelSizeYmm, patch, psfPixelSizeMm, status)
  !---------------------------------------------------------------------
  ! Requires cal to have been set up;
  !
!  ! Makes use of the global variable mmToArcsec
  !
  ! So, what is it with all this 'bogus' stuff? The true theta and phi
  ! (ie, coordinates in the TELCOORD system) of the sources are easily
  ! calculable. However if these values are used, the psf patch image comes
  ! out (naturally) in the true TELCOORD system, ie rotated with respect
  ! to the sky coordinates of the image. Next best is to more or less
  ! keep the theta value but calculate the equivalent to phi but in the
  ! sky coord system. The effect of this is that a PSF at the incorrect
  ! azimuth is used, but it appears with the correct orientation. To do
  ! it properly would require rebinning between rotated coordinate systems,
  ! which is much more difficult than the presently used rebinning between
  ! rotationally aligned systems. The PSF doesn't appear to vary greatly
  ! with azimuth in any case, so any bad effects are thought to be small.

    implicit none

    real(single), intent(in) ::&
      srcXpixels, srcYpixels,&
      fovCentreXpixels, fovCentreYpixels,&
      pixelSizeXmm, pixelSizeYmm
    real(double), intent(in) ::&
      energyFraction, EeV, srcThetaArcsec
    real(single), pointer :: patch(:,:)
    type(PsfBinSizeT), intent(out) :: psfPixelSizeMm
    integer, intent(out) :: status

    real(double) :: bogusPhi, thetaArcsec
    real(single) :: xMm, yMm
    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfPointer
    type(PsfWindowT) :: psfWindow

    !-------------------------------------------------------------------
    ! Start

    status = 0

    xMm = (srcXpixels - fovCentreXpixels) * pixelSizeXmm
    yMm = (srcYpixels - fovCentreYpixels) * pixelSizeYmm
    bogusPhi = atan2(-yMm, xMm)
    do while(bogusPhi < 0.0d0)
      bogusPhi = bogusPhi + TwoPi
    end do

    validPsfRanges = CAL_psfValidityRanges()

    if (EeV < validPsfRanges%eMin)                status = -1
    if (EeV > validPsfRanges%eMax)                status = -2
    if (srcThetaArcsec < validPsfRanges%thetaMin) status = -3
    if (srcThetaArcsec > validPsfRanges%thetaMax)&
    then
                                                  status = -4
      thetaArcsec = validPsfRanges%thetaMax * (1.0d0 - epsilon(thetaArcsec))
    else
      thetaArcsec = srcThetaArcsec
    end if
    if (bogusPhi < validPsfRanges%phiMin)         status = -5
    if (bogusPhi > validPsfRanges%phiMax)         status = -6

if (status < 0 .and. status .ne. -4) call error('','')
! Hence only non-zero status return possible is -4, when the theta is out
! of bounds. This we want to flag in the source list.

    psfPointer = CAL_getPSF(EeV, thetaArcsec, bogusPhi)
    psfPixelSizeMm = CAL_psfBinSize(psfPointer)

    psfWindow = CAL_psfEnboxedEnergy(psfPointer, fraction=energyFraction)

    call CAL_psfImage(psfPointer, psfWindow, patch)

    call vignetPatch(patch)

  end subroutine getCalPsfPatch

  !---------------------------------------------------------------------
  subroutine vignetPatch(patch)
  !---------------------------------------------------------------------
  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.

    implicit none

    real(single), intent(inout) :: patch(:,:)

    integer(int16) :: xsize, ysize
    integer(int16), pointer :: indices(:)
    real(single) ::&
      xHalfSize, yHalfSize,&
      xVector(size(patch, 1)), yVector(size(patch, 2))

    !-------------------------------------------------------------------
    ! Start

    xsize = size(patch, 1)
    xHalfSize = xsize / 2.0 - 0.5
    ysize = size(patch, 2)
    yHalfSize = ysize / 2.0 - 0.5

    indices => indicize(xsize)
    xVector = (indices - 1.0 - xHalfSize) / (xHalfSize + 1.0)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.
    xVector = 1.0 - xVector * xVector
    deallocate(indices)

    indices => indicize(ysize)
    yVector = (indices - 1.0 - yHalfSize) / (yHalfSize + 1.0)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.
    yVector = 1.0 - yVector * yVector
    deallocate(indices)

    patch = patch * spread(xVector, 2, ysize)&
                  * spread(yVector, 1, xsize)

  end subroutine vignetPatch

  !---------------------------------------------------------------------
  subroutine resizePatch(patch, srcXpixels, srcYpixels, psfPixelSizePixels)
  !---------------------------------------------------------------------

    implicit none

    real(single), pointer :: patch(:,:)
    real(single), intent(in) :: srcXpixels, srcYpixels
    type(PsfBinSizeT), intent(in) :: psfPixelSizePixels

    integer(int16) :: newStartXi, newStartYi, newFinisXi, newFinisYi,&
      oldXsize, oldYsize
    real(single), pointer :: newPatch(:,:)
    real(single) :: oldHalfSizeX, oldHalfSizeY,&
      oldLoX, oldHiX, newLoX, newHiX, oldLoY, oldHiY, newLoY, newHiY
    integer(int16), pointer :: indices(:)
    real(single), allocatable ::&
      oldXedges(:), oldYedges(:), newXedges(:), newYedges(:)
    integer :: status
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    oldXsize = size(patch, 1)
    oldYsize = size(patch, 2)

    oldHalfSizeX = (1 + ubound(patch, 1) - lbound(patch, 1)) / 2.0
    oldHalfSizeY = (1 + ubound(patch, 2) - lbound(patch, 2)) / 2.0
    oldLoX = srcXpixels - psfPixelSizePixels%binX * oldHalfSizeX
    oldHiX = srcXpixels + psfPixelSizePixels%binX * oldHalfSizeX
if (testFlag) then
  print *, 'ubound(patch, 1): ', ubound(patch, 1)
  print *, 'lbound(patch, 1): ', lbound(patch, 1)
  print *, 'oldHalfSizeX: ', oldHalfSizeX
  print *, 'psfPixelSizePixels%binX: ', psfPixelSizePixels%binX
  print *, 'srcXpixels: ', srcXpixels
  print *, 'oldLoX: ', oldLoX
  print *, 'oldHiX: ', oldHiX
  print *, ''
end if
    newStartXi = floor(oldLoX)
    newFinisXi = floor(oldHiX)
    newLoX = real(newStartXi)
    newHiX = real(newFinisXi + 1)
    oldLoY = srcYpixels - psfPixelSizePixels%binY * oldHalfSizeY
    oldHiY = srcYpixels + psfPixelSizePixels%binY * oldHalfSizeY
    newStartYi = floor(oldLoY)
    newFinisYi = floor(oldHiY)
    newLoY = real(newStartYi)
    newHiY = real(newFinisYi + 1)

    allocate(newPatch(newStartXi:newFinisXi,newStartYi:newFinisYi))
    allocate(oldXedges(size(   patch, 1)+1), oldYedges(size(   patch, 2)+1))
    allocate(newXedges(size(newPatch, 1)+1), newYedges(size(newPatch, 2)+1))

    indices => indicize(int(0, int16), oldXsize)
    oldXedges = indices / real(oldXsize)
if (testFlag) then
  print *, 'indices(1), indices(2): ', indices(1), indices(2)
  print *, 'real(oldXsize): ', real(oldXsize)
  print *, 'old oldXedges(1), oldXedges(2): ', oldXedges(1), oldXedges(2)
end if
    oldXedges = fracVal(oldLoX, oldHiX, oldXedges)
if (testFlag) then
  print *, 'oldLoX: ', oldLoX
  print *, 'oldHiX: ', oldHiX
  print *, 'new oldXedges(1), oldXedges(2): ', oldXedges(1), oldXedges(2)
end if
    deallocate(indices)

    indices => indicize(int(0, int16), oldYsize)
    oldYedges = indices / real(oldYsize)
    oldYedges = fracVal(oldLoY, oldHiY, oldYedges)
    deallocate(indices)

    indices => indicize(newStartXi, newFinisXi + int(1, int16))
    newXedges = real(indices, single)
    deallocate(indices)

    indices => indicize(newStartYi, newFinisYi + int(1, int16))
    newYedges = real(indices, single)
    deallocate(indices)

    call rebinImage(patch, oldXedges, oldYedges&
      , newPatch, newXedges, newYedges, status)
    if (status < 0) then
print *, status
call error('rebinStatus','')
    end if

    call CAL_releaseMemory(patch)
    patch => newPatch

  end subroutine resizePatch

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module simulations
